import numpy as np
import math
import scipy
from fractions import Fraction
import itertools
import biotuner
from biotuner.biotuner_utils import *
import matplotlib.pyplot as plt
from numpy import array, zeros, ones, arange, log2, sqrt, diff, concatenate
import pytuning
from math import gcd
from numpy import array, zeros, ones, arange, log2, sqrt, diff, concatenate
from scipy.stats import norm
from scipy.signal import argrelextrema, detrend
import scipy.signal as ss
from pytuning import create_euler_fokker_scale
from collections import Counter
from functools import reduce
from pytuning.utilities import normalize_interval
from pactools import Comodulogram, REFERENCES


'''---------------------------------------------------------Extended peaks-------------------------------------------------------------'''

'''EXTENDED PEAKS from expansions
'''

    
def EEG_harmonics_mult(peaks, n_harmonics, n_oct_up = 0):
    """
    Natural harmonics

    This function takes a list of frequency peaks as input and computes the desired number of harmonics
    with the formula: x, 2x, 3x ..., nx

    peaks: List (float)
        Peaks represent local maximum in a spectrum
    n_harmonics: int
        Number of harmonics to compute
    n_oct_up: int
        Defaults to 0. Corresponds to the number of octave the peaks are shifted 
    
    Returns
    -------
    multi_harmonics: array
        (n_peaks, n_harmonics + 1)
    """
    n_harmonics = n_harmonics + 2
    multi_harmonics = []
    multi_harmonics_rebound = []
    for p in peaks:
        multi_harmonics_r = []
        multi_harm_temp = []
        harmonics = []
        p = p * (2**n_oct_up)
        i = 1
        harm_temp = p
        while i < n_harmonics:
            harm_temp = p * i
            harmonics.append(harm_temp)
            i+=1
        multi_harmonics.append(harmonics)
    multi_harmonics = np.array(multi_harmonics)

    return multi_harmonics


def EEG_harmonics_div(peaks, n_harmonics, n_oct_up = 0, mode = 'div'):
    """
    Natural sub-harmonics

    This function takes a list of frequency peaks as input and computes the desired number of harmonics
    with using division: 

    peaks: List (float)
        Peaks represent local maximum in a spectrum
    n_harmonics: int
        Number of harmonics to compute
    n_oct_up: int
        Defaults to 0. Corresponds to the number of octave the peaks are shifted 
    mode: str
        Defaults to 'div'.
        'div': x, x/2, x/3 ..., x/n
        'div_add': x, (x+x/2), (x+x/3), ... (x+x/n)
        'div_sub': x, (x-x/2), (x-x/3), ... (x-x/n)
    Returns
    -------
    div_harmonics: array
        (n_peaks, n_harmonics + 1)
    div_harmonics_bounded: array
        (n_peaks, n_harmonics + 1)
    """
    n_harmonics = n_harmonics + 2
    div_harmonics = []
    for p in peaks:
        harmonics = []
        p = p * (2**n_oct_up)
        i = 1
        harm_temp = p
        while i < n_harmonics:
            if mode == 'div':
                harm_temp = (p/i)
            if mode == 'div_add':
                harm_temp = p + (p/i)
            if mode == 'div_sub':
                harm_temp = p - (p/i)
            harmonics.append(harm_temp)
            i+=1
        div_harmonics.append(harmonics)
    div_harmonics = np.array(div_harmonics)
    div_harmonics_bounded = div_harmonics.copy()
    #Rebound the result between 1 and 2
    for i in range(len(div_harmonics_bounded)):
        for j in range(len(div_harmonics_bounded[i])):
            div_harmonics_bounded[i][j] = rebound(div_harmonics_bounded[i][j])
    return div_harmonics, div_harmonics_bounded


def harmonic_fit(peaks, n_harm = 10, bounds = 1, function = 'mult', div_mode = 'div', n_common_harms = 5):
    """
    This function computes harmonics of a list of peaks and compares the lists of harmonics pairwise to find fitting
    between the harmonic series

    peaks: List (float)
        Peaks represent local maximum in a spectrum
    n_harm: int
        Number of harmonics to compute
    bounds: int
        Minimum distance (in Hz) between two frequencies to consider a fit
    function: str
        Defaults to 'mult'.
        'mult' will use natural harmonics
        'div' will use natural sub-harmonics
    div_mode: str
        Defaults to 'div'. See EEG_harmonics_div function.
        
    Returns
    -------
    
    
    """
    from itertools import combinations
    peak_bands = []
    for i in range(len(peaks)):
        peak_bands.append(i)
    if function == 'mult':
        multi_harmonics = EEG_harmonics_mult(peaks, n_harm)
    elif function == 'div':
        multi_harmonics, x = EEG_harmonics_div(peaks, n_harm, mode = div_mode)
    elif function == 'exp':
        multi_harmonics = []
        increments = []
        for h in range(n_harm+1):
            h += 1 
            multi_harmonics.append([i**h for i in peaks])
        multi_harmonics = np.array(multi_harmonics)
        multi_harmonics = np.moveaxis(multi_harmonics, 0, 1)
    #print(np.array(multi_harmonics).shape)
    list_peaks = list(combinations(peak_bands,2))
    #print(list_peaks)
    harm_temp = []
    harm_list1 = []
    harm_list2 = []
    harm_list = []
    harmonics = []
    for i in range(len(list_peaks)):
        harms, _, _, d, e, harm_list = compareLists(multi_harmonics[list_peaks[i][0]], multi_harmonics[list_peaks[i][1]], bounds)
        harm_temp.append(harms)
        harm_list1.append(d)
        harm_list2.append(e)
        harmonics.append(harm_list)
    harm_fit = np.array(harm_temp).squeeze()
    harmonics = reduce(lambda x, y: x+y, harmonics)
    most_common_harmonics= [h for h, h_count in Counter(harmonics).most_common(n_common_harms) if h_count > 1]
    harmonics = list(np.sort(list(set(harmonics))))
    if len(peak_bands) > 2:
        harm_fit = list(itertools.chain.from_iterable(harm_fit))
        harm_fit = [round(num, 3) for num in harm_fit]
        harm_fit = list(dict.fromkeys(harm_fit))
        harm_fit = list(set(harm_fit))
    return harm_fit, harm_list1, harm_list2, harmonics, most_common_harmonics

'''EXTENDED PEAKS from restrictions
'''

def consonance_peaks (peaks, limit):
    """
    This function computes consonance (for a given ratio a/b, when a < 2b, consonance corresponds to (a+b)/(a*b)) between peaks

    peaks: List (float)
        Peaks represent local maximum in a spectrum
    limit: float
        minimum consonance value to keep associated pairs of peaks
        
        Comparisons with familiar ratios:
        Unison-frequency ratio 1:1 yields a value of 2
        Octave-frequency ratio 2:1 yields a value of 1.5
        Perfect 5th-frequency ratio 3:2 yields a value of 0.833
        Perfect 4th-frequency ratio 4:3 yields a value of 0.583
        Major 6th-frequency ratio 5:3 yields a value of 0.533
        Major 3rd-frequency ratio 5:4 yields a value of 0.45
        Minor 3rd-frequency ratio 5:6 yields a value of 0.366
        Minor 6th-frequency ratio 5:8 yields a value of 0.325
        Major 2nd-frequency ratio 8:9 yields a value of 0.236
        Major 7th-frequency ratio 8:15 yields a value of 0.192
        Minor 7th-frequency ratio 9:16 yields a value of 0.174
        Minor 2nd-frequency ratio 15:16 yields a value of 0.129
        
    Returns
    -------
    
    consonance: List (float)
        consonance scores for each pairs of consonant peaks
    cons_pairs: List of lists (float)
        list of lists of each pairs of consonant peaks
    cons_peaks: List (float)
        list of consonant peaks (no doublons)
    cons_tot: float
        averaged consonance value for each pairs of peaks
        
    """
    from fractions import Fraction
    consonance_ = []
    peaks2keep = []
    peaks_consonance = []
    cons_tot = []
    for p1 in peaks:
        for p2 in peaks:
            peaks2keep_temp = []
            p2x = p2
            p1x = p1
            if p1x > p2x:
                while p1x > p2x:
                    p1x = p1x/2
            if p1x < p2x:
                while p2x > p1x:
                    p2x = p2x/2
            if p1x < 0.1:
                p1x = 0.06
            if p2x < 0.1:
                p2x = 0.06  #random  number to avoid division by 0
            ratio = Fraction(p2x/p1x).limit_denominator(1000)
            cons_ = (ratio.numerator + ratio.denominator)/(ratio.numerator * ratio.denominator)
            if cons_ < 1 :
                cons_tot.append(cons_)
            if cons_ > 1 or cons_ < limit:
                cons_ = None
                cons_ = None
                p2x = None
                p1x = None
            if p2x != None:
                peaks2keep_temp.extend([p2, p1])
            consonance_.append(cons_)
            peaks2keep.append(peaks2keep_temp)
        #cons_pairs = np.array(peaks2keep)
        cons_pairs = [x for x in peaks2keep if x]
        #consonance = np.array(consonance_)
        consonance = [i for i in consonance_ if i]
        cons_peaks = list(itertools.chain(*cons_pairs))
        cons_peaks = [np.round(c, 2) for c in cons_peaks]
        cons_peaks = list(set(cons_peaks))  
        #consonance = list(set(consonance))
    return consonance, cons_pairs, cons_peaks, np.average(cons_tot)


def multi_consonance(cons_pairs, n_freqs = 5):
    """
    Function that keeps the frequencies that are the most consonant with others
    Takes pairs of frequencies that are consonant (output of the 'compute consonance' function)

    cons_pairs: List of lists (float)
        list of lists of each pairs of consonant peaks
    n_freqs: int
        maximum number of consonant freqs to keep
    
    Returns
    -------
    
    freqs_related: List (float)
        peaks that are consonant with at least two other peaks, starting with the peak that is 
        consonant with the maximum number of other peaks
    """
    freqs_dup = list(itertools.chain(*cons_pairs))
    pairs_temp = list(itertools.chain.from_iterable(cons_pairs))
    freqs_nodup = list(dict.fromkeys(pairs_temp))
    f_count = []
    for f in freqs_nodup:
        f_count.append(freqs_dup.count(f))
    freqs_related = [x for _,x in sorted(zip(f_count,freqs_nodup))][-(n_freqs):][::-1]
    return freqs_related




def consonant_ratios (peaks, limit, sub = False, input_type = 'peaks', metric = 'cons'):
    """
    Function that computes integer ratios from peaks with higher consonance
    Needs at least two pairs of values
    
    peaks: List (float)
        Peaks represent local maximum in a spectrum
    limit: float
        minimum consonance value to keep associated pairs of peaks
    sub: boolean
        Defaults to False
        When set to True, include ratios a/b when a < b.
    
    Returns
    -------
    
    cons_ratios: List (float)
        list of consonant ratios
    consonance: List (float)
        list of associated consonance values
    """
    from fractions import Fraction
    consonance_ = []
    ratios2keep = []
    if input_type == 'peaks':
        ratios = compute_peak_ratios(peaks, sub = sub)
    if input_type == 'ratios':
        ratios = peaks
    for ratio in ratios:
        frac = Fraction(ratio).limit_denominator(1000)
        if metric == 'cons':
            cons_ = (frac.numerator + frac.denominator)/(frac.numerator * frac.denominator)
        if metric == 'harmsim':
            cons_ = dyad_similarity(ratio)
        if cons_ > limit :
            consonance_.append(cons_)
            ratios2keep.append(ratio)
    #print(ratios2keep)
    
    ratios2keep = np.array(np.round(ratios2keep, 3))
    cons_ratios = np.sort(list(set(ratios2keep)))
    #cons_ratios = np.array(ratios2keep)
    #ratios = []
    #ratios = [ratios.append(x) for x in ratios2keep if x not in ratios]
    consonance = np.array(consonance_)
    consonance = [i for i in consonance if i]
    return cons_ratios, consonance


def timepoint_consonance (data, method = 'cons', limit = 0.2, min_notes = 3):
    """
    ## Function that keeps moments of consonance from multiple time series of peak frequencies

    data: List of lists (float)
        Axis 0 represents moments in time
        Axis 1 represents the sets of frequencies
    method: str
        Defaults to 'cons'
        'cons' will compute pairwise consonance between frequency peaks in the form of (a+b)/(a*b)
        'euler' will compute Euler's gradus suavitatis
    limit: float
        limit of consonance under which the set of frequencies are not retained
        When method = 'cons'
             --> See consonance_peaks method's doc to refer consonance values to common intervals 
        When method = 'euler'
             --> Major (4:5:6) = 9
                 Minor (10:12:15) = 9
                 Major 7th (8:10:12:15) = 10
                 Minor 7th (10:12:15:18) = 11
                 Diminish (20:24:29) = 38 
    min_notes: int
        minimum number of consonant frequencies in the chords. Only relevant when method is set to 'cons'. 
    Returns
    -------
    
    chords: List of lists (float)
        Axis 0 represents moments in time
        Axis 1 represents the sets of consonant frequencies
    positions: List (int)
        positions on Axis 0
    """
    
    data = np.moveaxis(data, 0, 1)
    print('NAN', np.argwhere(np.isnan(data)))
    out = []
    positions = []
    for count, peaks in enumerate(data):
        peaks = [x for x in peaks if x >= 0]
        if method == 'cons':
            cons, b, peaks_cons, d = consonance_peaks(peaks, limit)
            out.append(peaks_cons)
            if len(list(set(peaks_cons))) >= min_notes:
                positions.append(count)
        if method == 'euler':
            peaks_ =  [int(np.round(p, 2)*100) for p in peaks]
            #print(peaks_)
            eul = euler(*peaks_)
            #print(eul)
            if eul < limit:
                out.append(list(peaks))
                positions.append(count)
    out = [x for x in out if x != []]
    #if method == 'cons':
    out = list(out for out,_ in itertools.groupby(out))
    chords = [x for x in out if len(x)>=min_notes]
    return chords, positions


'''
    ##################################################   PEAKS METRICS    ############################################################
'''

#Consonance#
#Input: peaks


def consonance (ratio, limit = 1000):
    ''' Compute metric of consonance from a single ratio of frequency
    
    ratio: float
    limit: int
        Defaults to 1000
        Maximum value of the denominator of the fraction representing the ratio  
    '''
    ratio = Fraction(float(ratio)).limit_denominator(limit)
    cons = (ratio.numerator + ratio.denominator)/(ratio.numerator * ratio.denominator)
    return cons

def euler(*numbers):
    """
    Euler's "gradus suavitatis" (degree of sweetness) function
    Return the "degree of sweetness" of a musical interval or chord expressed
    as a ratio of frequencies a:b:c, according to Euler's formula
    Greater values indicate more dissonance
    
    numbers: List (int)
        frequencies
    """
    factors = prime_factors(lcm(*reduced_form(*numbers)))
    return 1 + sum(p - 1 for p in factors)

#Input: peaks
def tenneyHeight(peaks, avg = True):
    """
    Tenney Height is a measure of inharmonicity calculated on two frequencies (a/b) reduced in their simplest form. 
    It can also be called the log product complexity of a given interval.
    
    peaks: List (float)
        frequencies
    avg: Boolean
        Default to True
        When set to True, all tenney heights are averaged
    """
    pairs = getPairs(peaks)
    pairs
    tenney = []
    for p in pairs:
        try:
            frac = Fraction(p[0]/p[1]).limit_denominator(1000)
        except ZeroDivisionError:
            p[1] = 0.01
            frac = Fraction(p[0]/p[1]).limit_denominator(1000)
        x = frac.numerator
        y = frac.denominator
        tenney.append(log2(x*y))
    if avg == True:
        tenney = np.average(tenney)
    return tenney


def peaks_to_metrics (peaks, n_harm = 10):
    '''
    This function computes different metrics on peak frequencies.
    
    peaks: List (float)
        Peaks represent local maximum in a spectrum
    n_harm: int
        Number of harmonics to compute for 'harm_fit' metric
        
    Returns
    -------
    
    metrics: dict (float)
        Dictionary of values associated to metrics names
    metrics_list: List (float)
        list of peaks metrics values in the order: 'cons', 'euler', 'tenney', 'harm_fit'  
    '''
    peaks = list(peaks)
    metrics = {'cons' : 0, 'euler' : 0, 'tenney': 0, 'harm_fit': 0}
    harm_fit, harm_pos1, harm_pos2 = harmonic_fit(peaks, n_harm = n_harm)
    metrics['harm_pos1'] = harm_pos1
    metrics['harm_pos2'] = harm_pos2
    
    metrics['harm_fit'] = len(harm_fit)

    a, b, c, metrics['cons'] = consonance_peaks (peaks, 0.1)
    peaks_highfreq  = [int(p*1000) for p in peaks]
    
    metrics['euler'] = euler(*peaks_highfreq)
    metrics['tenney'] = tenneyHeight(peaks_highfreq) 
    metrics_list = []
    for value in metrics.values():
        metrics_list.append(value)
    return metrics, metrics_list


def metric_denom(ratio):
    '''Function that computes the denominator of the normalized ratio
    ratio: float
    '''
    
    ratio = sp.Rational(ratio).limit_denominator(10000)
    normalized_degree = normalize_interval(ratio)
    y = int(sp.fraction(normalized_degree)[1])
    return y


'''SCALE METRICS'''
'''Metric of harmonic similarity represents the degree of similarity between a scale and the natural harmonic series ###
   Implemented from Gill and Purves (2009)'''

def dyad_similarity(ratio):
    '''
    This function computes the similarity between a dyad of frequencies and the natural harmonic series 
    ratio: float
        frequency ratio
    '''
    frac = Fraction(float(ratio)).limit_denominator(1000)
    x = frac.numerator
    y = frac.denominator
    z = ((x+y-1)/(x*y))*100
    return z

#Input: ratios (list of floats) 
def ratios2harmsim (ratios):
    '''
    This function computes the similarity for each ratio of a list
    ratios: List (float)
        list of frequency ratios (forming a scale)
        
    Returns
    ---------
    similarity: List (float)
        list of percentage of similarity for each ratios
    '''
    fracs = []
    for r in ratios:
        fracs.append(Fraction(r).limit_denominator(1000))
    sims = []
    for f in fracs:
        sims.append(dyad_similarity(f.numerator/f.denominator))
    similarity = np.array(sims)
    return similarity

def scale_cons_matrix (scale, function):
    '''
    This function gives a metric of a scale corresponding to the averaged metric for each pairs of ratios (matrix)
    scale: List (float)
    function: function
        possible functions: dyad_similarity
        consonance
        metric_denom
        
    '''
    metric_values = []
    mode_values = []
    for index1 in range(len(scale)):
        for index2 in range(len(scale)):
            if scale[index1] > scale[index2]:  #not include the diagonale in the computation of the avg. consonance
                entry = scale[index1]/scale[index2]
                #print(entry_value, scale[index1], scale[index2])
                mode_values.append([scale[index1], scale[index2]])
                metric_values.append(function(entry))
    return np.average(metric_values)

def PyTuning_metrics(scale, maxdenom):
    '''
    This function computes the scale metrics of the PyTuning library (https://pytuning.readthedocs.io/en/0.7.2/metrics.html)
    Smaller values are more consonant
    
    scale: List (float)
        List of ratios corresponding to scale steps
    maxdenom: int
        Maximum value of the denominator for each step's fraction
    
    '''
    scale_frac, num, denom = scale2frac(scale, maxdenom)
    metrics = pytuning.metrics.all_metrics(scale_frac)
    sum_p_q = metrics['sum_p_q']
    sum_distinct_intervals = metrics['sum_distinct_intervals']
    metric_3 = metrics['metric_3']
    sum_p_q_for_all_intervals = metrics['sum_p_q_for_all_intervals']
    sum_q_for_all_intervals = metrics['sum_q_for_all_intervals']
    return sum_p_q, sum_distinct_intervals, metric_3, sum_p_q_for_all_intervals, sum_q_for_all_intervals

def scale_to_metrics(scale):
    '''
    This function computes the scale metrics of the PyTuning library and other scale metrics
    
    scale: List (float)
        List of ratios corresponding to scale steps
        
    Returns
    ----------
    
    scale_metrics: dictionary
        keys correspond to metrics names
    scale_metrics_list: List (float)
        List of values corresponding to all computed metrics (in the same order as dictionary)
    '''
    scale_frac, num, denom = scale2frac(scale, maxdenom=1000)
    scale_metrics = pytuning.metrics.all_metrics(scale_frac)
    scale_metrics['harm_sim'] = np.round(np.average(ratios2harmsim(scale)), 2)
    scale_metrics['matrix_harm_sim'] = scale_cons_matrix(scale, dyad_similarity)
    scale_metrics['matrix_cons'] = scale_cons_matrix(scale, consonance)
    scale_metrics_list = []
    for value in scale_metrics.values():
        scale_metrics_list.append(value)
    return scale_metrics, scale_metrics_list


'''
    ################################################   SCALE CONSTRUCTION  ##############################################################

'''

def oct_subdiv(ratio, octave_limit = 0.01365 ,octave = 2 ,n = 5):
    '''
    N-TET tuning from Generator Interval
    This function uses a generator interval to suggest numbers of steps to divide the octave, 
    so the given interval will be approximately present (octave_limit) in the steps of the N-TET tuning.
    
    ratio: float
        ratio that corresponds to the generator_interval
        e.g.: by giving the fifth (3/2) as generator interval, this function will suggest to subdivide the octave in 12, 53, ...
    octave_limit: float 
        Defaults to 0.01365 (Pythagorean comma)
        approximation of the octave corresponding to the acceptable distance between the ratio of the generator interval after 
        multiple iterations and the octave value.
    octave: int
        Defaults to 2
        value of the octave
    n: int
        Defaults to 5
        number of suggested octave subdivisions
        
    Returns
    -------
    
    Octdiv: List (int)
        list of N-TET tunings corresponding to dividing the octave in equal steps
    Octvalue: List (float)
        list of the approximations of the octave for each N-TET tuning
    '''
    Octdiv, Octvalue, i = [], [], 1
    ratios = []
    while len(Octdiv) < n:
        ratio_mult = (ratio**i)
        while ratio_mult > octave:
            ratio_mult = ratio_mult/octave

        rescale_ratio = ratio_mult - round(ratio_mult)
        ratios.append(ratio_mult)
        i+=1
        if -octave_limit < rescale_ratio < octave_limit:
            Octdiv.append(i-1)
            Octvalue.append(ratio_mult)
        else:
            continue
    return Octdiv, Octvalue



def compare_oct_div(Octdiv = 12, Octdiv2 = 53, bounds = 0.005, octave = 2):
    '''
    Function that compare steps for two N-TET tunings and return matching ratios and corresponding degrees
    
    Octdiv: int
        Defaults to 12.
        first N-TET tuning number of steps
    Octdiv2: int
        Defaults to 53.
        second N-TET tuning number of steps
    bounds: float 
        Defaults to 0.005
        Maximum distance between 1 ratio of Octdiv and 1 ratio of Octdiv2 to consider a match
    octave: int
        Defaults to 2
        value of the octave
        
    Returns
    -------
    
    avg_ratios: List (float)
        list of ratios corresponding to the shared steps in the two N-TET tunings
    shared_steps: List of tuples
        the two elements of each tuple corresponds to the scale steps sharing the same interval in the two N-TET tunings
    '''
    ListOctdiv = []
    ListOctdiv2 = []
    OctdivSum = 1
    OctdivSum2 = 1
    i = 1
    i2 = 1
    while OctdivSum < octave:
        OctdivSum =(nth_root(octave, Octdiv))**i
        i+=1
        ListOctdiv.append(OctdivSum)
    while OctdivSum2 < octave:
        OctdivSum2 =(nth_root(octave, Octdiv2))**i2
        i2+=1
        ListOctdiv2.append(OctdivSum2)
    shared_steps = []
    avg_ratios = []
    for i, n in enumerate(ListOctdiv):
        for j, harm in enumerate(ListOctdiv2):
            if harm-bounds < n < harm+bounds:
                shared_steps.append((i+1, j+1))
                avg_ratios.append((n+harm)/2)
    return avg_ratios, shared_steps




#Output1: octave subdivisions
#Output2: ratios that led to Output1

def multi_oct_subdiv (peaks, max_sub = 100, octave_limit = 1.01365, octave = 2, n_scales = 10, cons_limit = 0.1):
    '''
    This function uses the most consonant peaks ratios as input of oct_subdiv function. Each consonant ratio
    leads to a list of possible octave subdivisions. These lists are compared and optimal octave subdivisions are
    determined.
    
    peaks: List (float)
        Peaks represent local maximum in a spectrum
    max_sub: int
        Defaults to 100.
        Maximum number of intervals in N-TET tuning suggestions.
    octave_limit: float 
        Defaults to 1.01365 (Pythagorean comma).
        Approximation of the octave corresponding to the acceptable distance between the ratio of the generator interval after 
        multiple iterations and the octave value.
    octave: int
        Defaults to 2.
        value of the octave
    n_scales: int
        Defaults to 10.
        Number of N-TET tunings to compute for each generator interval (ratio).
        
    Returns
    -------
    
    multi_oct_div: List (int)
        List of octave subdivisions that fit with multiple generator intervals.
    ratios: List (float)
        list of the generator intervals for which at least 1 N-TET tuning match with another generator interval.  
    '''
    import itertools
    from collections import Counter
    #a, b, pairs, cons = consonance_peaks(peaks, cons_limit)
    ratios, cons = consonant_ratios(peaks, cons_limit)
    list_oct_div = []
    for i in range(len(ratios)):
        list_temp, _ = oct_subdiv(ratios[i], octave_limit, octave, n_scales)
        list_oct_div.append(list_temp)
    counts = Counter(list(itertools.chain(*list_oct_div)))
    oct_div_temp = []
    for k, v in counts.items():
        if v > 1:
            oct_div_temp.append(k)
    oct_div_temp = np.sort(oct_div_temp)
    multi_oct_div = []
    for i in range(len(oct_div_temp)):
        if oct_div_temp[i] < max_sub:
            multi_oct_div.append(oct_div_temp[i])
    return multi_oct_div, ratios

def harmonic_tuning (list_harmonics, octave = 2, min_ratio = 1, max_ratio = 2):
    '''
    Function that computes a tuning based on a list of harmonic positions
    
    list_harmonics: List (int)
        harmonic positions to use in the scale construction
    octave: int
    min_ratio: float
    max_ratio: float
    '''
    ratios = []
    for i in list_harmonics:
        ratios.append(rebound(1*i, min_ratio, max_ratio, octave))
    ratios = list(set(ratios))
    ratios = list(np.sort(np.array(ratios)))
    return ratios

def euler_fokker_scale(intervals, n = 1):
    '''
    Function that takes as input a series of intervals and derives a Euler Fokker Genera scale
    intervals: List (float)
    n: int
        Defaults to 1
        number of times the interval is used in the scale generation
    '''
    multiplicities = [n for x in intervals]
    scale = create_euler_fokker_scale(intervals, multiplicities)
    return scale
    
def generator_interval_tuning (interval = 3/2, steps = 12, octave = 2):
    '''
    Function that takes a generator interval and derives a tuning based on its stacking.
    interval: float
        Generator interval
    steps: int
        Defaults to 12 (12-TET for interval 3/2)
        Number of steps in the scale
    octave: int
        Defaults to 2
        Value of the octave
    '''
    scale = []
    for s in range(steps):
        s += 1
        degree = interval**s
        while degree > octave:
            degree = degree/octave
        scale.append(degree)
    return sorted(scale)


#function that takes two ratios a input (boundaries of )
#The mediant corresponds to the interval where small and large steps are equal.

def tuning_range_to_MOS (frac1, frac2, octave = 2, max_denom_in = 100, max_denom_out = 100):
    gen1 = octave**(frac1)
    gen2 = octave**(frac2)
    a = Fraction(frac1).limit_denominator(max_denom_in).numerator
    b = Fraction(frac1).limit_denominator(max_denom_in).denominator
    c = Fraction(frac2).limit_denominator(max_denom_in).numerator
    d = Fraction(frac2).limit_denominator(max_denom_in).denominator
    print(a, b, c, d)
    mediant = (a+c)/(b+d)
    mediant_frac = sp.Rational((a+c)/(b+d)).limit_denominator(max_denom_out)
    gen_interval = octave**(mediant)
    gen_interval_frac = sp.Rational(octave**(mediant)).limit_denominator(max_denom_out)
    MOS_signature = [d, b]
    invert_MOS_signature = [b, d]
    
    
    return mediant, mediant_frac, gen_interval, gen_interval_frac, MOS_signature, invert_MOS_signature


#def tuning_embedding ()

def stern_brocot_to_generator_interval (ratio, octave = 2):
    gen_interval = octave**(ratio)
    return gen_interval


def gen_interval_to_stern_brocot (gen):
    root_ratio = log2(gen)
    return root_ratio


#Dissonance
def dissmeasure(fvec, amp, model='min'):
    """
    Given a list of partials in fvec, with amplitudes in amp, this routine
    calculates the dissonance by summing the roughness of every sine pair
    based on a model of Plomp-Levelt's roughness curve.
    The older model (model='product') was based on the product of the two
    amplitudes, but the newer model (model='min') is based on the minimum
    of the two amplitudes, since this matches the beat frequency amplitude.
    """
    # Sort by frequency
    sort_idx = np.argsort(fvec)
    am_sorted = np.asarray(amp)[sort_idx]
    fr_sorted = np.asarray(fvec)[sort_idx]

    # Used to stretch dissonance curve for different freqs:
    Dstar = 0.24  # Point of maximum dissonance
    S1 = 0.0207
    S2 = 18.96

    C1 = 5
    C2 = -5

    # Plomp-Levelt roughness curve:
    A1 = -3.51
    A2 = -5.75

    # Generate all combinations of frequency components
    idx = np.transpose(np.triu_indices(len(fr_sorted), 1))
    fr_pairs = fr_sorted[idx]
    am_pairs = am_sorted[idx]

    Fmin = fr_pairs[:, 0]
    S = Dstar / (S1 * Fmin + S2)
    Fdif = fr_pairs[:, 1] - fr_pairs[:, 0]

    if model == 'min':
        a = np.amin(am_pairs, axis=1)
    elif model == 'product':
        a = np.prod(am_pairs, axis=1)  # Older model
    else:
        raise ValueError('model should be "min" or "product"')
    SFdif = S * Fdif
    D = np.sum(a * (C1 * np.exp(A1 * SFdif) + C2 * np.exp(A2 * SFdif)))

    return D

#Input: peaks and amplitudes
def diss_curve (freqs, amps, denom=1000, max_ratio=2, euler_comp = True, method = 'min', plot = True, n_tet_grid = None):
    '''
    This function computes the dissonance curve and related metrics for a given set of frequencies (freqs) and amplitudes (amps)
    
    freqs: List (float)
        list of frequencies associated with spectral peaks
    amps: List (float)
        list of amplitudes associated with freqs (must be same lenght)
    denom: int
        Defaults to 1000.
        Highest value for the denominator of each interval
    max_ratio: int
        Defaults to 2.
        Value of the maximum ratio
        Set to 2 for a span of 1 octave
        Set to 4 for a span of 2 octaves
        Set to 8 for a span of 3 octaves
        Set to 2**n for a span of n octaves
    euler: Boolean
        Defaults to True
        When set to True, compute the Euler Gradus Suavitatis for the derived scale
    method: str
        Defaults to 'min'
        Can be set to 'min' or 'product'. Refer to dissmeasure function for more information.
    plot: boolean
        Defaults to True
        When set to True, a plot of the dissonance curve will be generated
    n_tet_grid: int
        Defaults to None
        When an integer is given, dotted lines will be add to the plot a steps of the given N-TET scale 
    
    Returns
    -------

    intervals: List of tuples
        Each tuple corresponds to the numerator and the denominator of each scale step ratio
    ratios: List (float)
        list of ratios that constitute the scale
    euler_score: int
        value of consonance of the scale
    diss: float
        value of averaged dissonance of the total curve
    dyad_sims: List (float)
        list of dyad similarities for each ratio of the scale
        
    '''
    from numpy import array, linspace, empty, concatenate
    from scipy.signal import argrelextrema
    from fractions import Fraction
    freqs = np.array(freqs)
    r_low = 1
    alpharange = max_ratio
    method = method
    n = 1000
    diss = empty(n)
    a = concatenate((amps, amps))
    for i, alpha in enumerate(linspace(r_low, alpharange, n)):
        f = concatenate((freqs, alpha*freqs))
        d = dissmeasure(f, a, method)
        diss[i] = d
    diss_minima = argrelextrema(diss, np.less)
    
    
    intervals = []
    for d in range(len(diss_minima[0])):        
        frac = Fraction(diss_minima[0][d]/(n/(max_ratio-1))+1).limit_denominator(denom)
        frac = (frac.numerator, frac.denominator)
        intervals.append(frac)
    intervals.append((2, 1))
    ratios = [i[0]/i[1] for i in intervals]
    ratios_sim = [np.round(r, 2) for r in ratios] #round ratios for similarity measures of harmonic series
    #print(ratios_sim)
    dyad_sims = ratios2harmsim(ratios[:-1]) # compute dyads similarities with natural harmonic series
    dyad_sims
    a = 1
    ratios_euler = [a]+ratios    
    ratios_euler = [int(round(num, 2)*1000) for num in ratios]
    print(ratios_euler)
    euler_score = None
    if euler_comp == True:
        euler_score = euler(*ratios_euler)
        
        euler_score = euler_score/len(diss_minima)
    else:
        euler_score = 'NaN'
    
    if plot == True:
        plt.figure(figsize=(14, 6))
        plt.plot(linspace(r_low, alpharange, len(diss)), diss)
        plt.xscale('linear')
        plt.xlim(r_low, alpharange)
        try:
            plt.text(1.9, 1.5, 'Euler = '+str(int(euler_score)), horizontalalignment = 'center',
        verticalalignment='center', fontsize = 16)
        except:
            pass
        for n, d in intervals:
            plt.axvline(n/d, color='silver')
        # Plot N-TET grid
        if n_tet_grid != None:
            n_tet = NTET_ratios(n_tet_grid, max_ratio = max_ratio)
        for n in n_tet :
            plt.axvline(n, color='red', linestyle = '--')
        # Plot scale ticks
        plt.minorticks_off()
        plt.xticks([n/d for n, d in intervals],
                   ['{}/{}'.format(n, d) for n, d in intervals], fontsize = 13)
        plt.yticks(fontsize = 13)
        plt.tight_layout()
        plt.show()
    return intervals, ratios, euler_score, np.average(diss), dyad_sims

'''Harmonic Entropy'''

def compute_harmonic_entropy_domain_integral(ratios, ratio_interval, spread=0.01, min_tol=1e-15):
    
    # The first step is to pre-sort the ratios to speed up computation
    ind = np.argsort(ratios)
    weight_ratios = ratios[ind]
    
    centers = (weight_ratios[:-1] + weight_ratios[1:]) / 2
    
    ratio_interval = array(ratio_interval)
    N = len(ratio_interval)
    HE = zeros(N)
    for i, x in enumerate(ratio_interval):
        P = diff(concatenate(([0], norm.cdf(log2(centers), loc=log2(x), scale=spread), [1])))
        ind = P > min_tol
        HE[i] = -np.sum(P[ind] * log2(P[ind]))
    
    return weight_ratios, HE

def compute_harmonic_entropy_simple_weights(numerators, denominators, ratio_interval, spread=0.01, min_tol=1e-15):
    
    # The first step is to pre-sort the ratios to speed up computation
    ratios = numerators / denominators
    ind = np.argsort(ratios)
    numerators = numerators[ind]
    denominators = denominators[ind]
    weight_ratios = ratios[ind]
    
    ratio_interval = array(ratio_interval)
    N = len(ratio_interval)
    HE = zeros(N)
    for i, x in enumerate(ratio_interval):
        P = norm.pdf(log2(weight_ratios), loc=log2(x), scale=spread) / sqrt(numerators * denominators)
        ind = P > min_tol
        P = P[ind]
        P /= np.sum(P)
        HE[i] = -np.sum(P * log2(P))
    
    return weight_ratios, HE


def harmonic_entropy (ratios, res = 0.001, spread = 0.01, plot_entropy = True, plot_tenney = False, octave = 2):
    '''
    Harmonic entropy is a measure of the uncertainty in pitch perception, and it provides a physical correlate of tonalness, 
    one aspect of the psychoacoustic concept of dissonance (Sethares). High tonalness corresponds to low entropy and low tonalness
    corresponds to high entropy.
    
    ratios: List (float)
        ratios between each pairs of frequency peaks
    res: float
        Defaults to 0.001
        resolution of the ratio steps
    spread: float
        Default to 0.01
    plot_entropy: boolean
        Defaults to True
        When set to True, plot the harmonic entropy curve
    plot_tenney: boolean
        Defaults to False
        When set to True, plot the tenney heights (y-axis) across ratios (x-axis)
    octave: int
        Defaults to 2
        Value of the maximum interval ratio
        
    Returns
    ----------
    HE_minima: List (float)
        List of ratios corresponding to minima of the harmonic entropy curve
    HE: float
        Value of the averaged harmonic entropy
        
    '''
    fracs, numerators, denominators = scale2frac(ratios)
    ratios = numerators / denominators
    #print(ratios)
    #ratios = np.interp(ratios, (ratios.min(), ratios.max()), (1, 10))
    bendetti_heights = numerators * denominators
    tenney_heights = log2(bendetti_heights)

    ind = np.argsort(tenney_heights)  # first, sort by Tenney height to make things more efficient
    bendetti_heights = bendetti_heights[ind]
    tenney_heights = tenney_heights[ind]
    numerators = numerators[ind]
    denominators = denominators[ind]
    #ratios = ratios[ind]
    if plot_tenney == True:
        fig = plt.figure(figsize=(10, 4), dpi=150)
        ax = fig.add_subplot(111)
        # ax.scatter(ratios, 2**tenney_heights, s=1)
        ax.scatter(ratios, tenney_heights, s=1, alpha=.2)
        # ax.scatter(ratios[:200], tenney_heights[:200], s=1, color='r')
        plt.show()
    
    # Next, we need to ensure a distance `d` between adjacent ratios
    M = len(bendetti_heights)

    delta = 0.00001
    indices = ones(M, dtype=bool)

    for i in range(M - 2):
        ind = abs(ratios[i + 1:] - ratios[i]) > delta
        indices[i + 1:] = indices[i + 1:] * ind

    bendetti_heights = bendetti_heights[indices]
    tenney_heights = tenney_heights[indices]
    numerators = numerators[indices]
    denominators = denominators[indices]
    ratios = ratios[indices]
    M = len(tenney_heights)
    #print(M)
    #print('hello')
    x_ratios = arange(1, octave, res)
    _, HE = compute_harmonic_entropy_domain_integral(ratios, x_ratios, spread=spread)
    #_, HE = compute_harmonic_entropy_simple_weights(numerators, denominators, x_ratios, spread=0.01)
    ind = argrelextrema(HE, np.less)
    HE_minima = (x_ratios[ind], HE[ind])
    if plot_entropy == True:  
        fig = plt.figure(figsize=(10, 4), dpi=150)
        ax = fig.add_subplot(111)
        # ax.plot(weight_ratios, log2(pdf))
        ax.plot(x_ratios, HE)
        # ax.plot(x_ratios, HE_simple)
        ax.scatter(HE_minima[0], HE_minima[1], color='k', s=4)
        ax.set_xlim(1, octave)
        plt.show()
    return HE_minima, np.average(HE)


'''Scale reduction'''

def scale_reduction (scale, mode_n_steps, function, rounding = 4):
    '''
    Function that reduces the number of steps in a scale according to the consonance between pairs of ratios
    scale: List (float)
        scale to reduce
    mode_n_steps: int
        number of steps of the reduced scale
    function: function
        function used to compute the consonance between pairs of ratios
        Choose between: consonance, dyad_similarity, metric_denom
    '''
    metric_values = []
    mode_values = []
    for index1 in range(len(scale)):
        for index2 in range(len(scale)):
            if scale[index1] > scale[index2]:  #not include the diagonale in the computation of the avg. consonance
                entry = scale[index1]/scale[index2]
                #print(entry_value, scale[index1], scale[index2])
                mode_values.append([scale[index1], scale[index2]])
                #if function == metric_denom:
                #    metric_values.append(int(function(sp.Rational(entry).limit_denominator(1000))))
                #else:
                metric_values.append(function(entry))
    if function == metric_denom:
        cons_ratios = [x for _, x in sorted(zip(metric_values, mode_values))]
    else:
        cons_ratios = [x for _, x in sorted(zip(metric_values, mode_values))][::-1]
    i = 0
    mode_ = []
    mode_out = []
    while len(mode_out) < mode_n_steps: 
        cons_temp = cons_ratios[i]  
        mode_.append(cons_temp)
        mode_out_temp = [item for sublist in mode_ for item in sublist]
        mode_out_temp = [np.round(x, rounding) for x in mode_out_temp]
        mode_out = sorted(set(mode_out_temp), key = mode_out_temp.index)[0:mode_n_steps]
        i +=1
    mode_metric = []
    for index1 in range(len(mode_out)):
        for index2 in range(len(mode_out)):
            if mode_out[index1] > mode_out[index2]:
                entry = mode_out[index1]/mode_out[index2]
                #if function == metric_denom:
                #    mode_metric.append(int(function(sp.Rational(entry).limit_denominator(1000))))
                #else:
                mode_metric.append(function(entry))
    return np.average(metric_values), mode_out, np.average(mode_metric)


'''------------------------------------------------------Peaks extraction--------------------------------------------------------------'''
import emd
from PyEMD import EMD, EEMD
from scipy.signal import butter, lfilter
import colorednoise as cn
#PEAKS FUNCTIONS
#HH1D_weightAVG (Hilbert-Huang 1D): takes the average of all the instantaneous frequencies weighted by power
#HH1D_max: takes the frequency bin that has the maximum power value 

def compute_peaks_ts (data, peaks_function = 'EMD', FREQ_BANDS = None, precision = 0.25, sf = 1000, max_freq = 80):
    alphaband = [[7, 12]]
    try:
        if FREQ_BANDS == None:
            FREQ_BANDS = [[2, 3.55], [3.55, 7.15], [7.15, 14.3], [14.3, 28.55], [28.55, 49.4]]
    except:
        pass
    if peaks_function == 'EEMD':
        IMFs = EMD_eeg(data)[1:6]
    if peaks_function == 'EMD':
        data = np.interp(data, (data.min(), data.max()), (0, +1))
        IMFs = emd.sift.sift(data)
        #IMFs = emd.sift.ensemble_sift(data)
        IMFs = np.moveaxis(IMFs, 0, 1)[1:6]
    try:
        peaks_temp = []
        amps_temp = []
        for imf in range(len(IMFs)):
            p, a = compute_peak(IMFs[imf], precision = precision, average = 'median')
            #print(p)
            peaks_temp.append(p)
            
            amps_temp.append(a)
            
        peaks_temp = np.flip(peaks_temp)
        amps_temp = np.flip(amps_temp)
    except:
        pass
    if peaks_function == 'HH1D_max':
        IMFs = EMD_eeg(data)
        IMFs = np.moveaxis(IMFs, 0, 1)
        IP, IF, IA = emd.spectra.frequency_transform(IMFs[:, 1:6], sf, 'nht')
        precision_hh = precision*2
        low = 1
        high = max_freq
        steps = int((high-low)/precision_hh)
        edges, bins = emd.spectra.define_hist_bins(low, high, steps, 'log')
        
        # Compute the 1d Hilbert-Huang transform (power over carrier frequency)
        spec = emd.spectra.hilberthuang_1d(IF, IA, edges)
        spec = np.moveaxis(spec, 0, 1)
        peaks_temp = []
        amps_temp = []
        for e, i in enumerate(spec):
            max_power = np.argmax(i)
            peaks_temp.append(bins[max_power])
            amps_temp.append(spec[e][max_power])
        peaks_temp = np.flip(peaks_temp)
        amps_temp = np.flip(amps_temp)
    #if peaks_function == 'HH1D_weightAVG':
        
    if peaks_function == 'adapt':
        p, a = compute_peaks_raw(data, alphaband, precision = precision, average = 'median')
        FREQ_BANDS = alpha2bands(p)
        peaks_temp, amps_temp = compute_peaks_raw(data, FREQ_BANDS, precision = precision, average = 'median')
    if peaks_function == 'fixed':
        peaks_temp, amps_temp = compute_peaks_raw(data, FREQ_BANDS, precision = precision, average = 'median')
    peaks = np.array(peaks_temp)
    amps = np.array(amps_temp)
    return peaks, amps



def extract_all_peaks (data, sf, precision, max_freq = None):
    if max_freq == None:
        max_freq = sf/2
    mult = 1/precision
    nperseg = sf*mult
    nfft = nperseg
    freqs, psd = scipy.signal.welch(data, sf, nfft = nfft, nperseg = nperseg, average = 'median')
    psd = 10. * np.log10(psd)
    indexes = ss.find_peaks(psd, height=None, threshold=None, distance=10, prominence=None, width=2, wlen=None, rel_height=0.5, plateau_size=None)
    peaks = []
    amps = []
    for i in indexes[0]:
        peaks.append(freqs[i])
        amps.append(psd[i])
    peaks = np.around(np.array(peaks), 5)
    peaks = list(peaks)
    peaks = [p for p in peaks if p<=max_freq]
    return peaks, amps

def harmonic_peaks_fit (peaks, amps, min_freq = 0.5, max_freq = 30, min_harms = 2, harm_limit = 128):
    n_total = []
    harm_ = []
    harm_peaks = []
    max_n = []
    max_peaks = []
    max_amps = []
    harmonics = []
    harmonic_peaks = []
    harm_peaks_fit = []
    for p, a in zip(peaks, amps):
        n = 0
        harm_temp = []
        harm_peaks_temp = []
        if p < max_freq and p > min_freq:
            
            for p2 in peaks:
                if p2 == p:
                    ratio = 0.1 #arbitrary value to set ratio value to non integer
                if p2 > p:
                    ratio = p2/p
                    harm = ratio
                if p2 < p:
                    ratio = p/p2    
                    harm = -ratio
                if ratio.is_integer():
                    n += 1

                    harm_temp.append(harm)

                    harm_peaks_temp.append(p)
                    harm_peaks_temp.append(p2)
        n_total.append(n)
        harm_.append(harm_temp)
        harm_peaks.append(list(set(harm_peaks_temp)))
        if n >= min_harms:
            max_n.append(n)
            max_peaks.append(p)
            max_amps.append(a)
            #print(harm_peaks)
            harmonics.append(harm_temp)
            harmonic_peaks.append(harm_peaks)
            harm_peaks_fit.append([p, harm_temp, list(set(harm_peaks_temp))])
    max_n = np.array(max_n)
    max_peaks = np.array(max_peaks)
    max_amps = np.array(max_amps)
    harmonics = np.array(harmonics)
    for l in range(len(harmonics)):
        harmonics[l] = [x for x in harmonics[l] if x < harm_limit]
    harmonic_peaks = np.array(harmonic_peaks)
    #harm_peaks_fit = np.array(harm_peaks_fit)

    #max_indexes = np.argsort(n_total)[-10:]
    
    return max_n, max_peaks, max_amps, harmonics, harmonic_peaks, harm_peaks_fit


def cepstrum(signal, sample_freq, plot_cepstrum = False):
    windowed_signal = signal
    dt = 1/sample_freq
    freq_vector = np.fft.rfftfreq(len(windowed_signal), d=dt)
    X = np.fft.rfft(windowed_signal)
    log_X = np.log(np.abs(X))

    cepstrum = np.fft.rfft(log_X)
    cepstrum = smooth(cepstrum, 10)
    df = freq_vector[1] - freq_vector[0]
    quefrency_vector = np.fft.rfftfreq(log_X.size, df)
    quefrency_vector = smooth(quefrency_vector, 10)

    if plot_cepstrum == True:
        fig, ax = plt.subplots()
        ax.plot(freq_vector, log_X)
        ax.set_xlabel('frequency (Hz)')
        ax.set_title('Fourier spectrum')
        ax.set_xlim(0, 50)
        fig, ax = plt.subplots()
        ax.plot(quefrency_vector, np.abs(cepstrum))
        ax.set_xlabel('quefrency (s)')
        ax.set_title('cepstrum')
        ax.set_xlim(0.02, 0.5)
        ax.set_ylim(0, 200)
    return cepstrum, quefrency_vector


def cepstral_peaks (cepstrum, quefrency_vector, max_time, min_time):

    indexes = ss.find_peaks(cepstrum, height=None, threshold=None, distance=None, prominence=None, width=3, wlen=None, rel_height=0.5,         plateau_size=None)
    #print(indexes[0])
    peaks = []
    amps = []
    for i in indexes[0]:
        if quefrency_vector[i] < max_time and quefrency_vector[i] > min_time:
            amps.append(np.abs(cepstrum)[i])
            peaks.append(quefrency_vector[i])
    peaks = np.around(np.array(peaks), 3)
    peaks = list(peaks)
    #peaks = [p for p in peaks if p<=max_freq]
    peaks = [1/p for p in peaks]
    return peaks, amps


'''--------------------------------------------------Phase-Amplitude Coupling-----------------------------------------------------------'''


def pac_frequencies(ts, sf, method = 'duprelatour', n_values = 10, drive_precision = 0.05, max_drive_freq = 6, min_drive_freq = 3
                   , sig_precision = 1, max_sig_freq = 50, min_sig_freq = 8, 
                   low_fq_width = 0.5, high_fq_width = 1, plot = False):
    
    drive_steps = int(((max_drive_freq-min_drive_freq)/drive_precision)+1)
    low_fq_range = np.linspace(min_drive_freq, max_drive_freq, drive_steps)   
    sig_steps = int(((max_sig_freq-min_sig_freq)/sig_precision)+1)
    high_fq_range = np.linspace(min_sig_freq, max_sig_freq, sig_steps)
    
    estimator = Comodulogram(fs=sf, low_fq_range=low_fq_range,
                             low_fq_width=low_fq_width, high_fq_width = high_fq_width, 
                             high_fq_range = high_fq_range, method=method,
                             progress_bar=False)
    estimator.fit(ts)
    indexes = top_n_indexes(estimator.comod_, n_values)[::-1]
    pac_freqs = []
    for i in indexes:
        pac_freqs.append([low_fq_range[i[0]], high_fq_range[i[1]]])
    
    if plot == True:
        estimator.plot(titles=[REFERENCES[method]])
    return pac_freqs



def pac_most_frequent(pac_freqs, n):
    drive_freqs = [x[0] for x in pac_freqs]
    signal_freqs = [x[1]for x in pac_freqs]

    drive_dict = {k: v for k, v in sorted(Counter(drive_freqs).items(), key=lambda item: item[1])}
    max_drive = list(drive_dict)[::-1][0:n]

    signal_dict = {k: v for k, v in sorted(Counter(signal_freqs).items(), key=lambda item: item[1])}
    max_signal = list(signal_dict)[::-1][0:n]
    
    return [max_signal, max_drive]

def pac_mode(pac_freqs, n, function = dyad_similarity):
    _, mode, _ = scale_reduction(scale_from_pairs(pac_freqs), mode_n_steps = n, function = function)
    return sorted(mode)


'''--------------------------------------------------------Biorhythms-----------------------------------------------------------------'''


def scale2euclid(scale, max_denom = 10, mode = 'normal'):
    euclid_patterns = []
    frac, num, denom = scale2frac(scale, maxdenom = max_denom)
    if mode == 'normal':
        for n, d in zip(num, denom):
            if d <= max_denom:
                try:
                    euclid_patterns.append(bjorklund(n, d))
                except:
                    pass
    if mode == 'full':
        for d, n in zip(num, denom):
            if d <= max_denom:
                steps = d*n
                try:
                    euclid_patterns.append(bjorklund(steps, d))
                    euclid_patterns.append(bjorklund(steps, n))
                except:
                    pass
    return euclid_patterns

def invert_ratio(ratio, n_steps_down, limit_denom = 64):
    inverted_ratio = 1/(ratio)
    i = 2
    if n_steps_down >= 1:
        while i <= n_steps_down:

            inverted_ratio = inverted_ratio/ratio
            i+=1

    
    frac = sp.Rational(inverted_ratio).limit_denominator(limit_denom)
    return frac, inverted_ratio

def binome2euclid(binome, n_steps_down = 1, limit_denom = 64):
    euclid_patterns = []
    fracs = []
    new_binome = []
    new_frac1, b1 = invert_ratio(binome[0], n_steps_down, limit_denom = limit_denom)
    new_frac2, b2 = invert_ratio(binome[1], n_steps_down, limit_denom = limit_denom)
    new_binome.append(b1)
    new_binome.append(b2)
    frac, num, denom = scale2frac(new_binome, limit_denom)
    if denom[0] != denom[1]:
        new_denom = denom[0]*denom[1]
        #print('denom', new_denom)
        #print('num1', num[0]*denom[1])
        #print('num2', num[1]*denom[0])
        try:
            euclid_patterns.append(bjorklund(new_denom, num[0]*denom[1]))
            euclid_patterns.append(bjorklund(new_denom, num[1]*denom[0]))
        except:
            pass
        
    else:
        new_denom = denom[0]
        
        try:
            euclid_patterns.append(bjorklund(new_denom, num[0]))
            euclid_patterns.append(bjorklund(new_denom, num[1]))
        except:
            pass
        
    return euclid_patterns, [new_frac1, new_frac2], [[num[0]*denom[1], new_denom], [num[1]*denom[0], new_denom]]


def consonant_euclid (scale, n_steps_down, limit_denom, limit_cons, limit_denom_final):
    
    pairs = getPairs(scale)
    new_steps = []
    euclid_final = []
    for p in pairs:
        euclid, fracs, new_ratios = binome2euclid(p, n_steps_down, limit_denom)
        #print('new_ratios', new_ratios)
        new_steps.append(new_ratios[0][1])
    pairs_steps = getPairs(new_steps)
    cons_steps = []
    for steps in pairs_steps:   
        #print(steps)
        try:
            steps1 = Fraction(steps[0]/steps[1]).limit_denominator(steps[1]).numerator
            steps2 = Fraction(steps[0]/steps[1]).limit_denominator(steps[1]).denominator
            #print(steps1, steps2)
            cons = (steps1 + steps2)/(steps1 * steps2)
            if cons >= limit_cons and steps[0] <= limit_denom_final and steps[1] <= limit_denom_final:
                cons_steps.append(steps[0])
                cons_steps.append(steps[1])
        except:
            continue
    for p in pairs:
        euclid, fracs, new_ratios = binome2euclid(p, n_steps_down, limit_denom)
        if new_ratios[0][1] in cons_steps:
            
            try:
                euclid_final.append(euclid[0])
                euclid_final.append(euclid[1]) #exception for when only one euclid has been computed (when limit_denom is very low, chances to have a division by zero)
            except:
                pass
    euclid_final = sorted(euclid_final)
    euclid_final = [euclid_final[i] for i in range(len(euclid_final)) if i == 0 or euclid_final[i] != euclid_final[i-1]]
    euclid_final = [i for i in euclid_final if len(Counter(i).keys()) != 1]
    return euclid_final, cons_steps

def interval_vector(euclid):
    indexes = [index+1 for index, char in enumerate(euclid) if char == 1]
    length = len(euclid)+1
    vector = [t - s for s, t in zip(indexes, indexes[1:])]
    vector = vector+[length-indexes[-1]]
    return vector

def bjorklund(steps, pulses):
    steps = int(steps)
    pulses = int(pulses)
    if pulses > steps:
        raise ValueError    
    pattern = []    
    counts = []
    remainders = []
    divisor = steps - pulses
    remainders.append(pulses)
    level = 0
    while True:
        counts.append(divisor // remainders[level])
        remainders.append(divisor % remainders[level])
        divisor = remainders[level]
        level = level + 1
        if remainders[level] <= 1:
            break
    counts.append(divisor)
    
    def build(level):
        if level == -1:
            pattern.append(0)
        elif level == -2:
            pattern.append(1)         
        else:
            for i in range(0, counts[level]):
                build(level - 1)
            if remainders[level] != 0:
                build(level - 2)
    
    build(level)
    i = pattern.index(1)
    pattern = pattern[i:] + pattern[0:i]
    return pattern

def interval_vec_to_string (interval_vectors):
    strings = []
    for i in interval_vectors:
        strings.append('E('+str(len(i))+','+str(sum(i))+')')
    return strings

def euclid_string_to_referent (strings, dict_rhythms):
    referent = []
    for s in strings:
        if s in dict_rhythms.keys():
            referent.append(dict_rhythms[s])
        else:
            referent.append('None')
    return referent



#!bin/bash
"""Collection of operations on timeseries."""
import numpy as np
from PyEMD import EMD, EEMD
from scipy.signal import butter, lfilter
import matplotlib.pyplot as plt
import seaborn as sbn
from scipy import stats
import pygame, pygame.sndarray
import pytuning
from pytuning import *
import scipy.signal
import sympy as sp
import functools
import itertools
import operator
import sys
import bottleneck as bn
sys.setrecursionlimit(120000)



def compute_peak_ratios(peaks, rebound=True, octave=2, sub=False):
    """
    Ratios between peaks.

    This function calculates all the ratios (with the possibility to bound them
    between 1 and 2) derived from input peaks.

    peaks: List (float)
        Peaks represent local maximum in a spectrum
    rebound: boolean
        Defaults to True. False will output unbounded ratios
    octave: int
        Arbitrary wanted number of octaves. Defaults to 2.
    sub: boolean
        Defaults to False. True will include ratios below the unison (1)
    """
    # Iterating through successive peaks
    ratios = []
    peak_ratios_final = []
    for p1 in peaks:
        for p2 in peaks:

            # If a peak of value '0' is present, we skip this ratio computation
            if p1 < 0.1:
                p1 = 0.1
            try:
                ratio_temp = p2/p1
            except:
                pass
                
            # When sub is set to False, only ratios with numerators higher than denominators are consider
            if sub is False:
                if ratio_temp < 1:
                    ratio_temp = None
            # Ratios of 1 correspond to a peak divided by itself and are therefore not considered
            if ratio_temp == 1:
                ratio_temp = None
            ratios.append(ratio_temp)

        # I imagine there's a reason why you reinit with array
        peak_ratios = np.array(ratios)
        peak_ratios = [i for i in peak_ratios if i]  # dealing with NaNs ?
        peak_ratios = sorted(list(set(peak_ratios)))
        ratios_final = peak_ratios.copy()
    
    # If rebound is given, all the ratios are constrained between the unison and the octave
    if rebound is True:
        for peak in peak_ratios:

            # will divide the ratio by the octave until it reaches a value under the octave
            if peak > octave:
                while peak > octave:
                    peak = peak/octave
                peak_ratios_final.append(peak)
            # will multiply the ratio by the octave until it reaches a value over the unison (1)
            if peak < octave:
                while peak < 1:
                    peak = peak*octave
                peak_ratios_final.append(peak)
        # Preparing output
        peak_ratios_final = np.array(peak_ratios_final)
        peak_ratios_final = [i for i in peak_ratios_final if i]
        ratios_final = sorted(list(set(peak_ratios_final)))
    

    return ratios_final


def rebound(x, low=1, high=2, octave=2):
    """
    Rescale x within given octave bounds.

    x: int
        represents a peak value
    low: int
        Lower bound. Defaults to 1. 
    high: int
        Higher bound. Defaults to 2.
    octave: int
        Value of an octave
    """
    while x >= high:
        x = x/octave
    while x <= low:
        x = x*octave
    return x

def rebound_list(x_list, low=1, high=2, octave=2):
    return [rebound(x, low, high, octave) for x in x_list] 

def nth_root (num, root):
    '''
    This function computes the ratio associated with one step of a N-TET scale
    
    num: int
        value of the octave
    root: int
        number of steps in the N-TET scale
    '''
    answer = num**(1/root)
    return answer

def NTET_steps (octave, step, NTET):
    '''
    This function computes the ratio associated with a specific step of a N-TET scale
    
    octave: int
        value of the octave
    step: int
        value of the step
    NTET: int
        number of steps in the N-TET scale
    '''
    answer = octave**(step/NTET)
    return answer

#Function that compares lists (i.e. peak harmonics)
def compareLists(list1, list2, bounds):
    matching = []
    matching_pos = []
    matching_pos1 = []
    matching_pos2 = []
    positions = []
    for i, l1 in enumerate(list1):
        for j, l2 in enumerate(list2):
            if l2-bounds < l1 < l2+bounds:
                matching.append((l1+l2)/2)
                matching_pos.append([(l1+l2)/2, i+1, j+1])
                matching_pos1.append([list1[0], i+1])
                matching_pos2.append([list2[0], j+1])
                positions.append(i+1)
                positions.append(j+1)
    matching = np.array(matching)
    matching_pos = np.array(matching_pos)
    ratios_temp = []
    for i in range(len(matching_pos)):
        if matching_pos[i][1]>matching_pos[i][2]:
            ratios_temp.append(matching_pos[i][1]/matching_pos[i][2])
            #print(matching_pos[i][0])
        else:
            ratios_temp.append(matching_pos[i][2]/matching_pos[i][1])
    matching_pos_ratios = np.array(ratios_temp)
    return matching, matching_pos, matching_pos_ratios, matching_pos1, matching_pos2, positions

def create_SCL(scale, fname):
    #Output SCL files
    from pytuning.scales.pythagorean import create_pythagorean_scale
    from pytuning.tuning_tables import create_scala_tuning
    #scale = create_pythagorean_scale()
    table = create_scala_tuning(scale,fname)
    outF = open(fname+'.scl', "w")
    outF.writelines(table)
    outF.close()
    return

from fractions import Fraction
from math import log
from functools import reduce

def gcd(*numbers):
    """
    Return the greatest common divisor of the given integers
    The result will have the same sign as the last number given (so that
    when the last number is divided by the result, the result comes out
    positive).
    """
    def gcd(a, b):
        while b:
            a, b = b, a % b
        return a

    return reduce(gcd, numbers)

def gcd_n(*args):
    i = 1
    x = args[0]

    while i < len(args):
        x = gcd(x, args[i])
        i += 1
    return x

def reduced_form(*numbers):
    """
    Return a tuple of numbers which is the reduced form of the input,
    which is a list of integers
    """
    return tuple(int(a // gcd(*numbers)) for a in numbers)
def lcm(*numbers):
    """
    Return the least common multiple of the given integers
    """
    def lcm(a, b):
        if a == b == 0:
            return 0
        return (a * b) // gcd(a, b)

    # LCM(a, b, c, d) = LCM(a, LCM(b, LCM(c, d)))
    return reduce(lcm, numbers)

def prime_factors(n):
    """
    Return a list of the prime factors of the integer n.
    Don't use this for big numbers; it's a dumb brute-force method.
    """
    factors = []
    lastresult = n

    while lastresult > 1:
        c = 2
        while lastresult % c > 0:
            c += 1
        factors.append(c)
        lastresult /= c

    return factors

def top_n_indexes(arr, n):
    idx = bn.argpartition(arr, arr.size-n, axis=None)[-n:]
    width = arr.shape[1]
    return [divmod(i, width) for i in idx]

def scale_from_pairs(pairs):   
    return[rebound((x[1]/x[0])) for x in pairs]

def EMD_eeg (eeg_data):
    s = np.interp(eeg_data, (eeg_data.min(), eeg_data.max()), (0, +1))
    eemd = EEMD()
    t = np.linspace(0, 1, len(eeg_data))
    # Say we want detect extrema using parabolic method
    emd = eemd.EMD
    emd.extrema_detection="parabol"
    eIMFs = EMD().emd(s,t)
    #eIMFs = eemd.eemd(S, t)
    nIMFs = eIMFs.shape[0]
    return eIMFs


def butter_bandpass(lowcut, highcut, fs, order=5):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a

def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y

from scipy.fftpack import rfft, irfft

def phaseScrambleTS(ts):
    """Returns a TS: original TS power is preserved; TS phase is shuffled."""
    fs = rfft(ts)
    # rfft returns real and imaginary components in adjacent elements of a real array
    pow_fs = fs[1:-1:2]**2 + fs[2::2]**2
    phase_fs = np.arctan2(fs[2::2], fs[1:-1:2])
    phase_fsr = phase_fs.copy()
    np.random.shuffle(phase_fsr)
    # use broadcasting and ravel to interleave the real and imaginary components.
    # The first and last elements in the fourier array don't have any phase information, and thus don't change
    fsrp = np.sqrt(pow_fs[:, np.newaxis]) * np.c_[np.cos(phase_fsr), np.sin(phase_fsr)]
    fsrp = np.r_[fs[0], fsrp.ravel(), fs[-1]]
    tsr = irfft(fsrp)
    return tsr


##Surrogate data testing using Pyunicorn
try:
    from pyunicorn.timeseries.surrogates import *
    from pyunicorn.timeseries import RecurrenceNetwork
except:
    pass
#Surrogates.silence_level = 1
def AAFT_surrogates(self, original_data):
    """
    Return surrogates using the amplitude adjusted Fourier transform
    method.

    Reference: [Schreiber2000]_

    :type original_data: 2D array [index, time]
    :arg original_data: The original time series.
    :rtype: 2D array [index, time]
    :return: The surrogate time series.
    """
    #  Create sorted Gaussian reference series
    gaussian = random.randn(original_data.shape[0], original_data.shape[1])
    gaussian.sort(axis=1)

    #  Rescale data to Gaussian distribution
    ranks = original_data.argsort(axis=1).argsort(axis=1)
    rescaled_data = np.zeros(original_data.shape)

    for i in range(original_data.shape[0]):
        rescaled_data[i, :] = gaussian[i, ranks[i, :]]

    #  Phase randomize rescaled data
    phase_randomized_data = \
        correlated_noise_surrogates(self, rescaled_data)

    #  Rescale back to amplitude distribution of original data
    sorted_original = original_data.copy()
    sorted_original.sort(axis=1)

    ranks = phase_randomized_data.argsort(axis=1).argsort(axis=1)

    for i in range(original_data.shape[0]):
        rescaled_data[i, :] = sorted_original[i, ranks[i, :]]

    return rescaled_data

def correlated_noise_surrogates(self, original_data):
    """
    Return Fourier surrogates.

    Generate surrogates by Fourier transforming the :attr:`original_data`
    time series (assumed to be real valued), randomizing the phases and
    then applying an inverse Fourier transform. Correlated noise surrogates
    share their power spectrum and autocorrelation function with the
    original_data time series.

    The Fast Fourier transforms of all time series are cached to facilitate
    a faster generation of several surrogates for each time series. Hence,
    :meth:`clear_cache` has to be called before generating surrogates from
    a different set of time series!

    .. note::
       The amplitudes are not adjusted here, i.e., the
       individual amplitude distributions are not conserved!

    **Examples:**

    The power spectrum is conserved up to small numerical deviations:

    >>> ts = Surrogates.SmallTestData().original_data
    >>> surrogates = Surrogates.\
            SmallTestData().correlated_noise_surrogates(ts)
    >>> all(r(np.abs(np.fft.fft(ts,         axis=1))[0,1:10]) == \
            r(np.abs(np.fft.fft(surrogates, axis=1))[0,1:10]))
    True

    However, the time series amplitude distributions differ:

    >>> all(np.histogram(ts[0,:])[0] == np.histogram(surrogates[0,:])[0])
    False

    :type original_data: 2D array [index, time]
    :arg original_data: The original time series.
    :rtype: 2D array [index, time]
    :return: The surrogate time series.
    """
    if self.silence_level <= 1:
        print("Generating correlated noise surrogates...")

    #  Calculate FFT of original_data time series
    #  The FFT of the original_data data has to be calculated only once,
    #  so it is stored in self._original_data_fft.
    #if self._fft_cached:
    #    surrogates = self._original_data_fft
    #else:
    surrogates = np.fft.rfft(original_data, axis=1)
    self._original_data_fft = surrogates
    self._fft_cached = True

    #  Get shapes
    (N, n_time) = original_data.shape
    len_phase = surrogates.shape[1]

    #  Generate random phases uniformly distributed in the
    #  interval [0, 2*Pi]
    phases = random.uniform(low=0, high=2 * np.pi, size=(N, len_phase))

    #  Add random phases uniformly distributed in the interval [0, 2*Pi]
    surrogates *= np.exp(1j * phases)

    #  Calculate IFFT and take the real part, the remaining imaginary part
    #  is due to numerical errors.
    return np.ascontiguousarray(np.real(np.fft.irfft(surrogates, n=n_time,
                                                     axis=1)))

## This function takes instantaneous frequencies (IF) from Hilbert-Huang Transform (HH) as an array in the form of [time, freqs]
## and outputs the average euler consonance for the whole time series (euler_tot) and the number of moments frequencies had consonance
## below euler threshold (euler_tresh)
def HH_cons (IF, euler_tresh = 100, mult = 10):
    euler_tot = []
    cons_ind = []
    euler_good = []
    for t in range(len(IF)):
        freqs = [int(np.round(x*mult)) for x in IF[t]]

        for i in freqs:
            if i <= 0:
                #print('negative')
                #print(IF[t])
                freqs = [1 for x in freqs]
        if mult > 10:
            if freqs[-1] > 10*mult or freqs[-2] > 15*mult or freqs[-3] > 20*mult:
                freqs = [1 for x in freqs]
                print('non-expected high freq detected')
        #print(freqs)
        e_temp = euler(*freqs)
        if e_temp > 1 and e_temp < euler_tresh:
            cons_ind.append(t)
        if e_temp > 1:
            euler_good.append(e_temp)
        euler_tot.append(e_temp)
    return euler_good, cons_ind, euler_tot

def getPairs(peaks):
    peaks_ = peaks.copy()
    out = []
    for i in range(len(peaks_)-1):
        a = peaks_.pop(i-i)
        for j in peaks_:
            out.append([a, j])
    return out

def functools_reduce(a):
    return functools.reduce(operator.concat, a)

def scale2frac (scale, maxdenom = 1000):
    num = []
    den = []
    scale_frac = []
    for step in scale:
        frac = Fraction(step).limit_denominator(maxdenom)
        frac_ = sp.Rational(step).limit_denominator(maxdenom)
        num.append(frac.numerator)
        den.append(frac.denominator)
        scale_frac.append(frac_)
    return scale_frac, np.array(num), np.array(den)

def sort_scale_by_consonance(scale):
    cons_tot = []
    for step in scale:
        frac = Fraction(step).limit_denominator(1000)
        cons = (frac.numerator + frac.denominator)/(frac.numerator * frac.denominator)
        cons_tot.append(cons)
    sorted_scale = list(np.flip([x for _,x in sorted(zip(cons_tot,scale))]))
    return sorted_scale

def ratio2frac (ratio, maxdenom = 1000):
    
    frac = Fraction(ratio).limit_denominator(maxdenom)
    num = frac.numerator
    den = frac.denominator
    return [num, den]

def frac2scale (scale):
    scale_ratio = []
    for step in scale:
        scale_ratio.append(float(step))
    return scale_ratio

def peaks_to_amps (peaks, freqs, amps, sf):
    bin_size = np.round((sf/2)/len(freqs), 3)
    amps_out = []
    #print(amps.shape)
    for p in peaks:
        #print(p)
        index = int(p/bin_size)
        amp = amps[index]
        amps_out.append(amp)
    return amps_out

def chords_to_ratios(chords, harm_limit = 2, spread = True):
    chords_ratios = []
    chords_ratios_bounded = []
    for chord in chords:
        chord = sorted(chord)
        if harm_limit != None:
            if spread == True: #will allow each note to be within the defined harm_limit of the previous note 
                for note in range(len(chord)):
                    while chord[note] > chord[note-1]*harm_limit:
                        chord[note] = chord[note]/2
            if spread == False: #will allow each note to be within the defined harm_limit of the first note
                for note in range(len(chord)):
                    while chord[note] > chord[0]*2:
                        chord[note] = chord[note]/2
        chord = sorted([np.round(n, 1) for n in chord])
        chord = [int(n*10) for n in chord]
        gcd_chord = 2 #arbitrary number that is higher than 1
        while gcd_chord > 1:
            gcd_chord = gcd_n(*chord)
            if gcd_chord > 1:
                chord = [int(note/gcd_chord) for note in chord]
        chord_bounded = [c/chord[0] for c in chord]
        chords_ratios_bounded.append(chord_bounded)
        chords_ratios.append(chord)
    return chords_ratios, chords_ratios_bounded

def NTET_ratios (n_steps, max_ratio):
    steps = []
    for s in range(n_steps):
        steps.append(2**(s/n_steps))
    steps_out = []
    for j in range(max_ratio-1):
        steps_out.append([i+j for i in steps])
    steps_out = sum(steps_out, [])
    return steps_out

def bjorklund(steps, pulses):
    '''From https://github.com/brianhouse/bjorklund'''
    steps = int(steps)
    pulses = int(pulses)
    if pulses > steps:
        raise ValueError    
    pattern = []    
    counts = []
    remainders = []
    divisor = steps - pulses
    remainders.append(pulses)
    level = 0
    while True:
        counts.append(divisor // remainders[level])
        remainders.append(divisor % remainders[level])
        divisor = remainders[level]
        level = level + 1
        if remainders[level] <= 1:
            break
    counts.append(divisor)
    
    def build(level):
        if level == -1:
            pattern.append(0)
        elif level == -2:
            pattern.append(1)         
        else:
            for i in range(0, counts[level]):
                build(level - 1)
            if remainders[level] != 0:
                build(level - 2)
    
    build(level)
    i = pattern.index(1)
    pattern = pattern[i:] + pattern[0:i]
    return pattern


'''Spectromorphology functions'''

import pyACA

def computeFeatureCl_new(afAudioData, cFeatureName, f_s, window=4000, overlap=1):

    # compute feature
    [v, t] = pyACA.computeFeature(cFeatureName, afAudioData, f_s, None, window, overlap)

    return (v, t)

def EMD_to_spectromorph (IMFs,  sf, method = "SpectralCentroid", window = None, overlap = 1, in_cut =None, out_cut = None):
    if in_cut == None:
        in_cut = int(sf/2)
    if out_cut == None:
        out_cut = int(len(IMFs[0])-(sf/2))
    if window == None:
        window = int(sf/2)
    spectro_IMF = []
    for e in IMFs:
        f, t = computeFeatureCl_new(e, method, sf, window, overlap)
        #[f,t] = pyACA.computePitch('TimeAcf', e, 1000, afWindow=None, iBlockLength=1000, iHopLength=200)
        #df[i] = np.squeeze(f)
        if method == 'SpectralCentroid':
            try:
                spectro_IMF.append(f[0][in_cut:out_cut])
            except:
                spectro_IMF.append(f[in_cut:out_cut])
        if method == 'SpectralFlux':
            spectro_IMF.append(f[in_cut:out_cut])
    spectro_IMF = np.array(spectro_IMF)
    return spectro_IMF


def butter_bandpass(lowcut,highcut,fs,order=8):
    nyq = 0.5*fs
    low = lowcut/nyq
    high = highcut/nyq

    b,a = butter(order, [low, high], btype='band')
    return b,a

def butter_bandpass_filter(data,lowcut,highcut,fs,order=8):
    b,a = butter_bandpass(lowcut,highcut,fs,order=order)
    return lfilter(b,a,data) 


def ratios_harmonics (ratios, n_harms = 1):
    ratios_harms = []
    for h in range(n_harms):
        h += 1 
        ratios_harms.append([i*h for i in ratios])
    ratios_harms = [i for sublist in ratios_harms for i in sublist]
    return ratios_harms

def ratios_increments (ratios, n_inc = 1):
    ratios_harms = []
    for h in range(n_inc):
        h += 1 
        ratios_harms.append([i**h for i in ratios])
    ratios_harms = [i for sublist in ratios_harms for i in sublist]
    ratios_harms = list(set(ratios_harms))
    return ratios_harms


def smooth(x,window_len=11,window='hanning'):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    #if x.ndim != 1:
        #raise ValueError, "smooth only accepts 1 dimension arrays."

    #if x.size < window_len:
        #raise ValueError, "Input vector needs to be bigger than window size."


    #if window_len<3:
        #return x


    #if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        #raise ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"


    s=np.r_[x[window_len-1:0:-1],x,x[-2:-window_len-1:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=np.ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')

    y=np.convolve(w/w.sum(),s,mode='valid')
    return y



sample_rate = 44100
pygame.init()
pygame.mixer.init(frequency=44100, size=-16, channels=2, buffer=512)
   
def sine_wave(hz, peak, n_samples=sample_rate):
    """Compute N samples of a sine wave with given frequency and peak amplitude.
       Defaults to one second.
    """
    length = sample_rate / float(hz)
    omega = np.pi * 2 / length
    xvalues = np.arange(int(length)) * omega
    onecycle = peak * np.sin(xvalues)
    return np.resize(onecycle, (n_samples,)).astype(np.int16)

def square_wave(hz, peak, duty_cycle=.5, n_samples=sample_rate):
    """Compute N samples of a sine wave with given frequency and peak amplitude.
       Defaults to one second.
    """
    t = np.linspace(0, 1, 500 * 440/hz, endpoint=False)
    wave = scipy.signal.square(2 * np.pi * 5 * t, duty=duty_cycle)
    wave = np.resize(wave, (n_samples,))
    return (peak / 2 * wave.astype(np.int16))

waveform=sine_wave
def make_chord(hz, ratios):
    """Make a chord based on a list of frequency ratios."""
    sampling = 2048
    chord = waveform(hz, sampling)
    for r in ratios[1:]:
        chord = sum([chord, sine_wave(hz * r / ratios[0], sampling)])
    return chord

def major_triad(hz):
    return make_chord(hz, [4, 5, 6])

def listen_scale (scale, fund, length):
    print('Scale:', scale)
    scale = [1]+scale
    for s in scale:
        freq = fund*s
        print(freq)
        note = make_chord(freq, [1])
        note = np.ascontiguousarray(np.vstack([note,note]).T)
        sound = pygame.sndarray.make_sound(note)
        sound.play(loops=0, maxtime=0, fade_ms=0)
        pygame.time.wait(int(sound.get_length() * length))
        
def listen_chords (chords, mult = 10, length = 500):
    #chords = np.around(chords, 3)
    print('Chords:', chords)

    for c in chords:
        c = [i*mult for i in c]
        chord = make_chord(c[0], c[1:])
        chord = np.ascontiguousarray(np.vstack([chord,chord]).T)
        sound = pygame.sndarray.make_sound(chord)
        sound.play(loops=0, maxtime=0, fade_ms=0)
        pygame.time.wait(int(sound.get_length() * length))
        
def horogram_tree_steps (ratio1, ratio2, steps = 10, limit = 1000):
    ratios_list = [ratio1, ratio2]
    s = 0
    while s < steps:
        ratio3 = horogram_tree(ratio1, ratio2, limit)
        ratios_list.append(ratio3)
        ratio1 = ratio2
        ratio2 = ratio3
        s += 1
    frac_list = [ratio2frac(x) for x in ratios_list]
    return frac_list, ratios_list
        

def horogram_tree (ratio1, ratio2, limit):
    a = Fraction(ratio1).limit_denominator(limit).numerator
    b = Fraction(ratio1).limit_denominator(limit).denominator
    c = Fraction(ratio2).limit_denominator(limit).numerator
    d = Fraction(ratio2).limit_denominator(limit).denominator
    next_step = (a+c)/(b+d)
    return next_step
        
def phi_convergent_point (ratio1, ratio2):
    Phi = (1 + 5 ** 0.5) / 2
    a = Fraction(ratio1).limit_denominator(1000).numerator
    b = Fraction(ratio1).limit_denominator(1000).denominator
    c = Fraction(ratio2).limit_denominator(1000).numerator
    d = Fraction(ratio2).limit_denominator(1000).denominator
    convergent_point = (c*Phi+a)/(d*Phi+b)
    return convergent_point

def Stern_Brocot(n,a=0,b=1,c=1,d=1):
    if(a+b+c+d>n):
        return 0
    x=Stern_Brocot(n,a+c,b+d,c,d)
    y=Stern_Brocot(n,a,b,a+c,b+d)
    if(x==0):
        if(y==0):
            return [a+c,b+d]
        else:
            return [a+c]+[b+d]+y
    else:
        if(y==0):
            return [a+c]+[b+d]+x
        else:
            return [a+c]+[b+d]+x+y
        
def scale_interval_names(scale, reduce = False):
    try:
        type = scale[0].dtype == 'float64'
        if type == True:
            scale, _, _ = scale2frac(scale)   
    except:
        pass
    interval_names = []
    for step in scale:
        name = pytuning.utilities.ratio_to_name(step)
        if reduce == True and name != None:
            interval_names.append([step, name])
        if reduce == False:
            interval_names.append([step, name])
    return interval_names

''' Continued fractions '''

import math

def contFrac(x, k):
    cf = []
    q = math.floor(x)
    cf.append(q)
    x = x - q
    i = 0
    while x != 0 and i < k:
        q = math.floor(1 / x)
        cf.append(q)
        x = 1 / x - q
        i = i + 1
    return cf

def convergents (interval):
    value = log2(interval)
    convergents = list(contfrac.convergents(value))
    return convergents



dict_rhythms = {
  "E(2,3)" : "A common Afro-Cuban drum pattern when started on the second onset as in [101]. For example, it is the conga rhythm of the (6/8)-time Swing Tumbao. It is common in Latin American music, as for example in the Cueca, and the coros de clave. It is common in Arabic music, as for example in the Al Taer rhythm of Nubia. It is also a rhythmic pattern of the Drum Dance of the Slavey Indians of Northern Canada",
  "E(2,5)" : "A rhythm found in Greece, Namibia, Rwanda and Central Africa. It is also a 13th century Persian rhythm called Khafif-e-ramal, as well as the rhythm of the dance Makedonka from the FYROM. Tchaikovsky used it as the metric pattern in the second movement of his Symphony No. 6. Started on the second onset as in [10010] it is a rhythm found in Central Africa, Bulgaria, Turkey, Turkestan and Norway. It is also the metric pattern of Dave Brubeck's Take Five, as well as Mars from The Planets by Gustav Holst. Rotated as in [10010100100], it is a Serbian rhythmic pattern. When it is started on the fourth (last) onset it is the Daasa al kbiri rhythmic pattern of Yemen.",
  "E(3,4)" : "It is the archetypal pattern of the Cumbia from Colombia, as well as a Calypso rhythm from Trinidad. It is also a thirteenth century Persian rhythm called Khalif-e-saghil, as well as the trochoid choreic rhythmic pattern of ancient Greece.",
  "E(3,5)" : "When started on the second onset, it is a thirteenth century Persian rhythm by the name of Khafif-e-ramal, as well as a Rumanian folk-dance rhythm.",
  "E(3,7)" : "It is the Ruchenitza rhythm used in a Bulgarian folk-dance. It is also the metric pattern of Pink Floyd's Money.",
  "E(3,8)" : "It is none other than one of the most famous rythms on the planet. In Cuba it goes by the name of the tresillo and in the USA is often called the Habanera rhythm used in hundreds of rockabilly songs during the 1950s. It can often be heard in early rock-and-roll hits in the left-hand patterns of the piano, or played on the string bass or saxophone. A good example is the bass rhythm in Elvis Presley's Hound Dog. The tresillo pattern is also found widely in West African traditional music. For example, it is played on the atoke bell in the Sohu, an Ewe dance from Ghana. The tresillo can also be recognized as the first bar of the ubiquitous two-bar clave Son given by [1001001000101000].",
  "E(4,7)" : "It is a Ruchenitza Bulgarian folk-dance rhythm.",
  "E(4,9)" : "It is the Aksak rhythm of Turkey. It is also the metric pattern used by Dave Brubeck in his piece Rondo a la Turk.",
  "E(4,11)" : "It is the metric pattern used by Frank Zappa in his piece titled Outside Now.",
  "E(4,12)" : "It is periodic with four repetitions of E(1,3) = [100]. It is the (12/8)-time Fandago clapping pattern in the Flamenco music of southern Spain, where 1 denotes a loud clap and 0 soft clap.",
  "E(4,15)" : "It is the metric pattern of the pancam savari tal of North Indian music.",
  "E(5,6)" : "It yields the York-Samai pattern, a popular Arabic rhythm. It is also a handclapping rhythm used in the Al Medemi songs of Oman.",
  "E(5,7)" : "It is the Nawakhat pattern, a popular Arabic rhythm. In Nubia it is called the Al Noht rhythm.",
  "E(5,8)" : "It is the Cuban cinquillo pattern, the Malfuf rhythmic pattern of Egypt, as well as the Korean Nong Pyon drum pattern. Started on the second onset, it is a popular Middle Eastern rhythm, as well as the Timini rhythm of Senegal, the Adzogbo dance rhythm of Benin, the Spanish Tango, the Maksum of Egypt, and a 13th century Persian rhythm, the Al-saghil-alsani. When it is started on the third onset it is the Musemmen rhythm of Turkey. When it is started on the fourth onset it is the Kromanti rhythm of Surinam.",
  "E(5,9)" : "It is a popular Arabic rhythm called Agsag-Samai. Started on the second onset, it is a drum pattern used by the Venda in South Africa, as well as a Rumanian folk-dance rhythm. It is also the rhythmic pattern of the Sigaktistos rhythm of Greece, and the Samai aktsak rhythm of Turkey. Started on the third onset, it is the rhythmic pattern of the Nawahiid rhythm of Turkey.",
  "E(5,11)" : "It is the metric pattern of the Savari tala used in the Hindustani music of India. It is also a rhythmic pattern used in Bulgaria and Serbia. In Bulgaria is is used in the Kopanitsa. This metric pattern has been used by Moussorgsky in Pictures at an Exhibition. Started on the third onset, it is the rhythm of the dance Kalajdzijsko Oro from the FYROM, and it appears in Bulgarian music as well.",
  "E(5,12)" : "It is a common rhythm played in the Central African Republic by the Aka Pygmies. It is also the Venda clapping pattern of a South African children's song, and a rhythm pattern used in the FYROM. Started on the second onset, it is the Columbia bell pattern popular in Cuba and West Africa, as well as a drumming pattern used in the Chakacha dance of Kenya, and also used in the FYROM . Started on the third onset, it is the Bemba bell pattern used in Northern Zimbabwe, and the rhythm of the dance from the FYROM Ibraim Odza Oro. Started on the fourth onset, it is the Fume Fume bell pattern popular in West Africa, and is a rhythm used in the former Yugoslavia. Finally, when started on the fifth onset it is the Salve bell pattern used in the Dominican Republic in a rhythm called Canto de Vela in honor of the Virgin Mary, as well as the drum rhythmic pattern of the Moroccan Al Kudam.",
  "E(5,13)" : "It is a rhythm from the FYROM which is also played by starting it on the fourth onset as follows: [1010010010100].",
  "E(5,16)" : "It is the Bossa-Nova rhythm necklace of Brazil. The actual Bossa-Nova rhythm usually starts on the third onset as follows: [1001001000100100]. However, other starting places are also documented in world music practices, such as [1001001001000100].",
  "E(6,7)" : "It is the Pontakos rhythm of Greece when started on the sixth (last) onset.",
  "E(6,13)" : "It is the rhythm of the dance Mama Cone pita from the FYROM. Started on the third onset, it is the rhythm of the dance Postupano Oro from the FYROM, as well as the Krivo Plovdivsko Horo of Bulgaria.",
  "E(7,8)" : "When started on the seventh (last) onset, is a typical rhythm played on the Bendir (frame drum), and used in the accompaniment of songs of the Tuareg people of Libya.",
  "E(7,9)" : "It is the Bazaragana rhythmic pattern of Greece.",
  "E(7,10)" : "It is the Lenk fahhte rhythmic pattern of Turkey.",
  "E(7,12)" : "It is a common West African bell pattern. For example, it is used in the Mpre rhythm of the Ashanti people of Ghana. Started on the seventh (last) onset, it is a Yoruba bell pattern of Nigeria, a Babenzele pattern of Central Africa, and a Mende pattern of Sierra Leone.",
  "E(7,15)" : "It is a Bulgarian rhythm when started on the third onset.",
  "E(7,16)" : "It is a Samba rhythm necklace from Brazil. The actual Samba rhythm is [1010010101001010] obtained by starting E(7, 16) on the last onset, and it coincides with a rhythm from the FYROM. When E(7, 16) is started on the fifth onset it is a clapping pattern from Ghana. When it is started on the second onset it is a rhythmic pattern found in the former Yugoslavia.",
  "E(7,17)" : "It is a rhythm from the FYROM when started on the second onset.",
  "E(7,18)" : "It is a Bulgarian rhythmic pattern.",
  "E(8,17)" : "It is a Bulgarian rhythmic pattern which is also started on the fifth onset.",
  "E(8,19)" : "It is a Bulgarian rhythmic pattern when started on the second onset.",
  "E(9,14)" : "When started on the second onset, it is the rhythmic pattern of the Tsofyan rhythm of Algeria.",
  "E(9,16)" : "It is a rhythm necklace used in the Central African Republic. When it is started on the second onset it is a bell pattern of the Luba people of Congo. When it is started on the fourth onset it is a rhythm played in West and Central Africa, as well as a cow-bell pattern in the Brazilian samba. When it is started on the penultimate onset it is the bell pattern of the Ngbaka-Maibo rhythms of the Central African Republic.",
  "E(9,22)" : "It is a Bulgarian rhythmic pattern when started on the second onset.",
  "E(9,23)" : "It is a Bulgarian rhythm.",
  "E(11,12)" : "When started on the second onset, is thedrum pattern of the Rahmani (a cylindrical double-headed drum) used in the Sot silam dance from Mirbat in the South of Oman.",
  "E(11,24)" : "It is a rhythm necklace of the Aka Pygmies of Central Africa. It is usually started on the seventh onset. Started on the second onset, it is a Bulgarian rhythm.",
  "E(13,24)" : "It is another rhythm necklace of the Aka Pygmies of the upper Sangha. Started on the penultimate onset, it is the Bobangi metal-blade pattern used by the Aka Pygmies.",
  "E(15,34)" : "It is a Bulgarian rhythmic pattern when started on the penultimate onset."
}


import biotuner
from biotuner.biotuner_functions import *
from biotuner.biotuner_utils import *
from fooof import FOOOF
from fooof.sim.gen import gen_power_spectrum
from fooof.sim.utils import set_random_seed
from fooof.plts.spectra import plot_spectrum
from fooof.plts.annotate import plot_annotated_model
from fooof.bands import Bands
from fooof.analysis import get_band_peak_fm, get_band_peak_fg
import scipy.signal
from pytuning import create_euler_fokker_scale

class biotuner(object):
    
    '''Class used to derive peaks information, musical scales and related metrics from time series  
    
    Example of use:       
    biotuning = biotuner(sf = 1000)
    biotuning.peaks_extraction(data)
    biotuning.peaks_extension()
    biotuning.peaks_metrics()
    '''
    
    def __init__(self, sf, data = None, peaks_function = 'EEMD', precision = 0.1, compute_sub_ratios = False, 
                 n_harm = 10, harm_function = 'mult', extension_method = 'consonant_harmonic_fit',
                 ratios_n_harms = 5, ratios_harms = False, ratios_inc = True, ratios_inc_fit = False,
                scale_cons_limit = 0.1):
        '''
        sf: int
            sampling frequency (in Hz)
            
        ///// PEAKS EXTRACTION ARGUMENTS ///// 
        peaks_function: str
            Defaults to 'EEMD'.
            Defines the method to use for peak extraction
            Peaks values are defined as the frequency value of the FFT bin with maximum amplitude 
            'fixed' : ranges of frequency bands are fixed
            'adapt' : ranges of frequency bands are defined based on the alpha peak
            'EMD': Intrinsic Mode Functions (IMFs) are derived from Empirical Mode Decomposition (EMD)  
                   FFT is computed on each IMF
            'EEMD': Intrinsic Mode Functions (IMFs) are derived from Ensemble Empirical Mode Decomposition (EMD)  
                    FFT is computed on each IMF
            'HH1D_max': maximum values of the 1d Hilbert-Huang transform on each IMF using EEMD.
            'harmonic_peaks': keeps peaks for which a maximum of other peaks are harmonics
            'cepstrum': peak frequencies of the cepstrum (inverse Fourier transform (IFT) of the logarithm of the estimated signal spectrum)
            'FOOOF' : peaks rising above the aperiodic component
        precision: float
            Defaults to 0.1
            precision of the peaks (in Hz)
            When HH1D_max is used, bins are in log scale.
        compute_sub_ratios: str
            Default to False
            When set to True, include ratios < 1 in peaks_ratios attribute
            
        ///// EXTENDED PEAKS ARGUMENTS /////
        n_harm: int
            Defaults to 10.
            Set the number of harmonics to compute in harmonic_fit function
        harm_function: str
            Defaults to 'mult'
            Computes harmonics from iterative multiplication (x, 2x, 3x, ...nx) or division (x, x/2, x/3, ...x/n)
            Set to 'mult' or 'div'  
        extension_method: str
            ['harmonic_fit', 'consonant', 'multi_consonant', 'consonant_harmonic_fit', 'multi_consonant_harmonic_fit']
            
        ///// RATIOS EXTENSION ARGUMENTS /////
        ratios_n_harms: int
            Defaults to 5.
            Defines to number of harmonics or exponents for extended ratios
        ratios_harms: boolean
            Defaults to False.
            When set to True, harmonics (x*1, x*2, x*3...,x*n) of specified ratios will be computed.
        ratios_inc: boolean
            Defaults to True.
            When set to True, exponentials (x**1, x**2, x**3,...x**n) of specified ratios will be computed.
        ratios_inc_fit: boolean
            Defaults to False.
            When set to True, a fit between exponentials (x**1, x**2, x**3,...x**n) of specified ratios will be computed.
        
        '''
        '''Initializing data'''
        if type(data) != type(None):
            self.data = data
        self.sf = sf
        '''Initializing arguments for peak extraction'''
        self.peaks_function = peaks_function
        self.precision = precision
        self.compute_sub_ratios = compute_sub_ratios
        '''Initializing arguments for peaks metrics'''
        self.n_harm = n_harm
        self.harm_function = harm_function
        self.extension_method = extension_method
        '''Initializing dictionary for scales metrics'''
        self.scale_metrics = {}
        self.scale_cons_limit = scale_cons_limit
        '''Initializing arguments for ratios extension'''
        self.ratios_n_harms = ratios_n_harms
        self.ratios_harms = ratios_harms
        self.ratios_inc = ratios_inc
        self.ratios_inc_fit = ratios_inc_fit
        
    
    
    '''First method to use. Requires data as input argument
       Generates self.peaks and self.peaks_ratios attributes'''

    def peaks_extraction (self, data, peaks_function = None, FREQ_BANDS = None, precision = None, sf = None, min_freq = 1, max_freq = 60,
                          min_harms = 2, compute_sub_ratios = None, ratios_extension = False, ratios_n_harms = None, scale_cons_limit =
                          None, octave = 2, harm_limit = 128):
        '''
        
        Arguments
        -------------
        data: 1d array (float)
            biosignal to analyse
            
        peaks_function: refer to __init__
        
        compute_sub_ratios: Boolean
            If set to True, will include peaks ratios (x/y) when x < y
        
        FREQ_BANDS: List of lists (float)
            Each list within the list of lists sets the lower and upper limit of a frequency band
        
        precision: float
            Defaults to None
            precision of the peaks (in Hz)
            When HH1D_max is used, bins are in log scale.
            
        min_freq: float
            Defaults to 1
            minimum frequency value to be considered as a peak
            Used with 'harmonic_peaks' and 'HH1D_max' peaks functions
            
        max_freq: float
            Defaults to 60
            maximum frequency value to be considered as a peak
            Used with 'harmonic_peaks' and 'HH1D_max' peaks functions
            
        min_harms: int
            Defaults to 2
            minimum number of harmonics to consider a peak frequency using the 'harmonic_peaks' function
            
        ratios_extension: Boolean
            Defaults to False
            When set to True, peaks_ratios harmonics and increments are computed
        
        ratios_n_harms: int
            Defaults to None
            number of harmonics or increments to use in ratios_extension method
        
        scale_cons_limit: float
            Defaults to None
            minimal value of consonance to be reach for a peaks ratio to be included in the peaks_ratios_cons attribute
        
        octave: float
            Defaults to 2
            value of the octave
            
        harm_limit: int
            Default to 128
            maximum harmonic position to keep when the 'harmonic_peaks' method is used
            
        Attributes
        -------------
        self.peaks: List (float)
            List of frequency peaks
        self.amps: List (float)
            List of peaks amplitude
        self.peaks_ratios: List (float)
            List of ratios between all pairs of peaks
        self.peaks_ratios_cons: List (float)
            List of consonant peaks ratios 
        ----------If ratios_extension = True:----------
        self.peaks_ratios_harm: List (float)
            List of peaks ratios and their harmonics
        self.peaks_ratios_inc: List (float)
            List of peaks ratios and their increments (ratios**n)
        self.peaks_ratios_inc_bound: List (float)
            List of peaks ratios and their increments (ratios**n) bounded within one octave
        self.peaks_ratios_inc_fit: List (float)
            List of peaks ratios and their increments (ratios**n) 
        '''
        
        self.data = data
        if sf == None:
            sf = self.sf
        if precision == None:
            precision = self.precision
        if peaks_function == None:
            peaks_function = self.peaks_function
        if compute_sub_ratios == None:
            compute_sub_ratios = self.compute_sub_ratios
        if scale_cons_limit == None:
            scale_cons_limit = self.scale_cons_limit
        if ratios_n_harms == None:
            ratios_n_harms = self.ratios_n_harms
        self.peaks, self.amps = self.compute_peaks_ts (data, peaks_function = peaks_function, FREQ_BANDS = FREQ_BANDS, precision = precision, 
                                                       sf = sf, min_freq = min_freq, max_freq = max_freq, min_harms = min_harms, harm_limit = harm_limit)


        self.peaks_ratios = compute_peak_ratios(self.peaks, rebound = True, octave = octave, sub = compute_sub_ratios)
        
        self.peaks_ratios_cons, b = consonant_ratios (self.peaks, limit = scale_cons_limit)
        if ratios_extension == True:
            a, b, c = self.ratios_extension(self.peaks_ratios, ratios_n_harms = ratios_n_harms)
            if a != None:
                self.peaks_ratios_harms = a
            if b != None:
                self.peaks_ratios_inc = b
                self.peaks_ratios_inc_bound = [rebound(x, low = 1, high = octave, octave = octave) for x in b]
            if c != None:
                self.peaks_ratios_inc_fit = c
    
    '''Generates self.extended_peaks and self.extended_peaks_ratios attributes'''
    
    def peaks_extension (self, peaks = None, n_harm = None, method = None, harm_function = 'mult', div_mode = 'add',
                         cons_limit = 0.1, ratios_extension = False, harm_bounds = 1, scale_cons_limit = None):
        if peaks == None:
            peaks = self.peaks
        if n_harm == None:
            n_harm = self.n_harm
        if method == None:
            method = self.extension_method
        if scale_cons_limit == None:
            scale_cons_limit = self.scale_cons_limit
        if method == 'harmonic_fit':
            extended_peaks, _, _, harmonics, _ = harmonic_fit(peaks, n_harm, function = harm_function, div_mode = div_mode, bounds = harm_bounds)
            self.extended_peaks = np.sort(list(self.peaks)+list(set(extended_peaks)))
        if method == 'consonant':
            consonance, cons_pairs, cons_peaks, cons_metric = consonance_peaks (peaks, limit = cons_limit)
            self.extended_peaks = np.sort(np.round(cons_peaks, 3))
        if method == 'multi_consonant':
            consonance, cons_pairs, cons_peaks, cons_metric = consonance_peaks (peaks, limit = cons_limit)
            self.extended_peaks = np.sort(np.round(multi_consonance(cons_pairs, n_freqs = 10), 3))
        if method == 'consonant_harmonic_fit':
            extended_peaks, _, _, harmonics, _ = harmonic_fit(peaks, n_harm, function = harm_function, div_mode = div_mode, bounds = harm_bounds)
            consonance, cons_pairs, cons_peaks, cons_metric = consonance_peaks (extended_peaks, limit = cons_limit)
            self.extended_peaks = np.sort(np.round(cons_peaks, 3))    
        if method == 'multi_consonant_harmonic_fit':
            extended_peaks, _, _, harmonics, _ = harmonic_fit(peaks, n_harm, function = harm_function, div_mode = div_mode, bounds = harm_bounds)
            consonance, cons_pairs, cons_peaks, cons_metric = consonance_peaks (extended_peaks, limit = cons_limit)
            self.extended_peaks = np.sort(np.round(multi_consonance(cons_pairs, n_freqs = 10), 3))
        self.extended_peaks = [i for i in self.extended_peaks if i<self.sf/2]
        self.extended_amps = peaks_to_amps(self.extended_peaks, self.freqs, self.psd, self.sf)
        if len(self.extended_peaks) > 0:
            self.extended_peaks_ratios = compute_peak_ratios(self.extended_peaks, rebound = True)
            if ratios_extension == True:
                a, b, c = self.ratios_extension(self.extended_peaks_ratios)
                if a != None:
                    self.extended_peaks_ratios_harms = a
                if b != None:
                    self.extended_peaks_ratios_inc = b
                if c != None:
                    self.extended_peaks_ratios_inc_fit = c
            self.extended_peaks_ratios = [np.round(r, 2) for r in self.extended_peaks_ratios]
            self.extended_peaks_ratios = list(set(self.extended_peaks_ratios))
            self.extended_peaks_ratios_cons, b = consonant_ratios (self.extended_peaks, scale_cons_limit, sub = False)
        return self.extended_peaks, self.extended_peaks_ratios
    
    def ratios_extension (self, ratios, ratio_fit_bounds = 0.001, ratios_n_harms = None):
        if ratios_n_harms == None:
            ratios_n_harms = self.ratios_n_harms
        if self.ratios_harms == True:
            ratios_harms_ = ratios_harmonics(ratios, ratios_n_harms)
        else: 
            ratios_harms_ = None
        if self.ratios_inc == True:
            ratios_inc_ = ratios_increments(ratios, ratios_n_harms)
        else: 
            ratios_inc_ = None
        if self.ratios_inc_fit == True:
            ratios_inc_fit_, _, _, ratios_inc_fit_pos, _ = harmonic_fit(ratios, ratios_n_harms, function = 'exp', bounds = ratio_fit_bounds)
        else: 
            ratios_inc_fit_ = None
        return ratios_harms_, ratios_inc_, ratios_inc_fit_
    
    def compute_spectromorph (self, IMFs = None, sf = None, method = 'SpectralCentroid', window = None, overlap = 1, comp_chords = False, min_notes = 3, 
                              cons_limit = 0.2, cons_chord_method = 'cons', graph = False):
        if IMFs == None:
            try:
                IMFs = self.IMFs
            except:
                IMFs = EMD_eeg(self.data)[1:6]
                self.IMFs = IMFs
                
        if sf == None:
            sf = self.sf
        if window == None:
            window = int(sf/2)
            
        spectro_EMD = EMD_to_spectromorph(IMFs, 1000, method = method, window = window, overlap = overlap)
        self.spectro_EMD = np.round(spectro_EMD, 1)
        if method == 'SpectralCentroid':
            self.SpectralCentroid = self.spectro_EMD
        if method == 'SpectralFlux':
            self.SpectralFlux = self.spectro_EMD
        if comp_chords == True:
            self.spectro_chords, spectro_chord_pos = timepoint_consonance(self.spectro_EMD, method = cons_chord_method, 
                                                                           limit = cons_limit, min_notes = min_notes) 
            self.spectro_chords = [l[::-1] for l in self.spectro_chords]
        if graph == True:
            data = np.moveaxis(self.spectro_EMD, 0, 1)
            ax = sbn.lineplot(data=data[10:-10, :], dashes = False)
            print('2')
            ax.set(xlabel='Time Windows', ylabel=method)
            ax.set_yscale('log')
            plt.legend(scatterpoints=1, frameon=True, labelspacing=1, title='EMDs', loc = 'best')
            labels = ['EMD1', 'EMD2', 'EMD3', 'EMD4', 'EMD5', 'EMD6']
            for xc in spectro_chord_pos:
                plt.axvline(x=xc, c='black', linestyle = 'dotted')
            plt.show()
            
    def compute_peaks_metrics (self, n_harm = None, bounds = 1, harm_bounds = 1):
        if n_harm == None:
            n_harm = self.n_harm
            
        peaks = list(self.peaks)        
        metrics = {'cons' : 0, 'euler' : 0, 'tenney': 0, 'harm_fit': 0, 'harmsim':0}   
        harm_fit, _, _, harm_pos, common_harm_pos = harmonic_fit(peaks, n_harm = n_harm, bounds = harm_bounds)
        metrics['harm_pos'] = harm_pos
        metrics['common_harm_pos'] = common_harm_pos
        metrics['harm_fit'] = len(harm_fit)
        a, b, c, metrics['cons'] = consonance_peaks (peaks, 0.1)
        peaks_euler = [int(round(num, 2)*1000) for num in peaks]
        
        if self.peaks_function != 'cepstrum' and self.peaks_function != 'FOOOF' and self.peaks_function != 'harmonic_peaks':
            try:

                metrics['euler'] = euler(*peaks_euler)

            except:
                pass
        metrics['tenney'] = tenneyHeight(peaks)
        metrics['harmsim'] = np.average(ratios2harmsim(self.peaks_ratios[:-1]))
        if self.peaks_function == 'harmonic_peaks':
            metrics['n_harmonic_peaks'] = self.n_harmonic_peaks
        metrics_list = []
        for value in metrics.values():
            metrics_list.append(value)
        self.peaks_metrics_list = metrics_list
        self.peaks_metrics = metrics

    '''Methods to compute scales from whether peaks or extended peaks'''
    
    def compute_diss_curve (self, input_type = 'peaks', denom=1000, max_ratio=2, euler_comp = False, method = 'min', 
                            plot = False, n_tet_grid = 12, scale_cons_limit = None):
        if input_type == 'peaks':
            peaks = self.peaks
            amps = self.amps
        if input_type == 'extended_peaks':
            peaks = self.extended_peaks
            amps = self.extended_amps
        if scale_cons_limit == None:
            scale_cons_limit = self.scale_cons_limit
 
        peaks = [p*128 for p in peaks]
        amps = np.interp(amps, (np.array(amps).min(), np.array(amps).max()), (0.2, 0.8))
        
        intervals, self.diss_scale, euler_diss, diss, harm_sim_diss = diss_curve (peaks, amps, denom=denom,
                                                                                  max_ratio=max_ratio, euler_comp = euler_comp,
                                                                                  method = method, plot = plot, n_tet_grid = n_tet_grid)
        self.diss_scale_cons, b = consonant_ratios (self.diss_scale, scale_cons_limit, sub = False, input_type = 'ratios')      
        print('intervals', intervals)
        self.scale_metrics['diss_euler'] = euler_diss
        self.scale_metrics['dissonance'] = diss
        self.scale_metrics['diss_harm_sim'] = np.average(harm_sim_diss)
        self.scale_metrics['diss_n_steps'] = len(self.diss_scale)
        
    def compute_harmonic_entropy(self, input_type = 'peaks', res = 0.001, spread = 0.01, 
                                 plot_entropy = True, plot_tenney = False, octave = 2, rebound = True, sub = False,
                                scale_cons_limit = None):
        if input_type == 'peaks':
            ratios = compute_peak_ratios(self.peaks, rebound = rebound, sub = sub)
        if input_type == 'extended_peaks':
            ratios = compute_peak_ratios(self.extended_peaks, rebound = rebound, sub = sub)
        if input_type == 'extended_ratios_harms':
            ratios = self.extended_peaks_ratios_harms
        if input_type == 'extended_ratios_inc':
            ratios = self.extended_peaks_ratios_inc
        if input_type == 'extended_ratios_inc_fit':
            ratios = self.extended_peaks_ratios_inc_fit
        if scale_cons_limit == None:
            scale_cons_limit = self.scale_cons_limit
            
        HE_scale, HE = harmonic_entropy(ratios, res = res, spread = spread, plot_entropy = plot_entropy, plot_tenney = plot_tenney, octave = octave)
        self.HE_scale = HE_scale[0]
        self.HE_scale_cons, b = consonant_ratios (self.HE_scale, scale_cons_limit, sub = False, input_type = 'ratios')  
        self.scale_metrics['HE'] = HE
        self.scale_metrics['HE_n_steps'] = len(self.HE_scale)  
        self.scale_metrics['HE_harm_sim'] = np.average(ratios2harmsim(list(self.HE_scale)))
        '''
        ratios_euler = [a]+ratios
        ratios_euler = [int(round(num, 2)*1000) for num in ratios]
        euler_score = None
        if consonance == True:
            euler_score = euler(*ratios_euler)
            euler_score = euler_score/len(diss_minima)
        '''
    
    
    def euler_fokker_scale(self, intervals, octave = 2):
        multiplicities = [1 for x in intervals]
        scale = create_euler_fokker_scale(intervals, multiplicities, octave)
        self.euler_fokker = scale
        return scale
    
    def harmonic_tuning(self, list_harmonics, octave = 2, min_ratio = 1, max_ratio = 2):
        ratios = []
        for i in list_harmonics:
            ratios.append(rebound(1*i, min_ratio, max_ratio, octave))
        ratios = list(set(ratios))
        ratios = list(np.sort(np.array(ratios)))
        self.harmonic_tuning_ = ratios
        return ratios
    
    def harmonic_fit_tuning(self, n_harm = 128, bounds = 0.1, n_common_harms = 128):
        
        _, _, _, harmonics, common_harmonics = harmonic_fit(self.peaks, n_harm =n_harm, 
                                                            bounds = bounds, n_common_harms = n_common_harms)
        self.harmonic_fit_tuning_ = harmonic_tuning(common_harmonics)
        return self.harmonic_fit_tuning_
    
    def pac(self, sf=None, method = 'duprelatour', n_values = 10, drive_precision = 0.05, max_drive_freq = 6, min_drive_freq = 3
                   , sig_precision = 1, max_sig_freq = 50, min_sig_freq = 8, 
                   low_fq_width = 0.5, high_fq_width = 1, plot = False):
        if sf==None:
            sf = self.sf
        self.pac_freqs = pac_frequencies(self.data, sf, method = method, n_values = n_values , drive_precision = drive_precision,
                                         max_drive_freq = max_drive_freq,min_drive_freq = min_drive_freq , sig_precision = sig_precision, 
                                         max_sig_freq = max_sig_freq, min_sig_freq =min_sig_freq, low_fq_width = low_fq_width, high_fq_width
                                         = high_fq_width, plot = plot)
        return self.pac_freqs
    '''Methods called by the peaks_extraction method'''
    
    def compute_peak(self, eeg_data, sf=1000, nperseg = 0, nfft = 0, precision = 0.25, average = 'median'):
        if nperseg == 0:
            mult = 1/precision
            nperseg = sf*mult
            nfft = nperseg
        import scipy
        freqs, psd = scipy.signal.welch(eeg_data, sf, nfft = nfft, nperseg = nperseg, average = average)
        self.freqs = freqs
        self.psd = psd
        psd = 10. * np.log10(psd) 
        bin_size = (sf/2)/len(freqs)
        #min_index = int(minf/bin_size)
        #max_index = int(maxf/bin_size)
        index_max = np.argmax(np.array(psd))
        FREQS = np.array(freqs[index_max])
        amps = np.array(psd[index_max])

        return FREQS, amps
    
    def compute_peaks_raw(self, eeg_data, FREQ_BANDS, sf=1000, nperseg = 0, nfft = 0, precision=0.25, average = 'median'):
        if nperseg == 0:
            mult = 1/precision
            nperseg = sf*mult
            nfft = nperseg
        import scipy
        psd_all = []
        freqs_all = []
        FREQs_temp= []
        amp_temp = []

        for minf, maxf in FREQ_BANDS:
            freqs, psd = scipy.signal.welch(eeg_data, sf, nfft = nfft, nperseg = nperseg, average = average)
            self.freqs = freqs
            psd = 10. * np.log10(psd)
            self.psd = psd
            bin_size = (sf/2)/len(freqs)
            self.bin_size = bin_size
            min_index = int(minf/bin_size)
            max_index = int(maxf/bin_size)
            index_max = np.argmax(np.array(psd[min_index:max_index]))
             #   print(index_max) # Should not be zero in all bands (would signify strong 1/f trend)
            FREQs_temp.append(freqs[min_index+index_max])
            amp_temp.append(psd[min_index+index_max])

        FREQS = np.array(FREQs_temp)
        amps = np.array(amp_temp)
        return FREQS, amps
    
    def extract_all_peaks (self, data, sf, precision, max_freq = None):
        if max_freq == None:
            max_freq = sf/2
        mult = 1/precision
        nperseg = sf*mult
        nfft = nperseg
        freqs, psd = scipy.signal.welch(data, sf, nfft = nfft, nperseg = nperseg, average = 'median')
        psd = 10. * np.log10(psd)
        self.freqs = freqs
        self.psd = psd
        indexes = ss.find_peaks(psd, height=None, threshold=None, distance=10, prominence=None, width=2, wlen=None, rel_height=0.5, plateau_size=None)
        peaks = []
        amps = []
        for i in indexes[0]:
            peaks.append(freqs[i])
            amps.append(psd[i])
        peaks = np.around(np.array(peaks), 5)
        peaks = list(peaks)
        peaks = [p for p in peaks if p<=max_freq]
        return peaks, amps

    def compute_peaks_ts (self, data, peaks_function = 'EMD', FREQ_BANDS = None, precision = 0.25, sf = 1000, min_freq = 1, max_freq = 80, min_harms = 2, harm_limit = 128):
        alphaband = [[7, 12]]
        try:
            if FREQ_BANDS == None:
                FREQ_BANDS = [[2, 3.55], [3.55, 7.15], [7.15, 14.3], [14.3, 28.55], [28.55, 49.4]]
        except:
            pass
        if peaks_function == 'EEMD':
            IMFs = EMD_eeg(data)[1:6]
            self.IMFs = IMFs
        if peaks_function == 'EMD':
            data = np.interp(data, (data.min(), data.max()), (0, +1))
            IMFs = emd.sift.sift(data)
            #IMFs = emd.sift.ensemble_sift(data)
            IMFs = np.moveaxis(IMFs, 0, 1)[1:6]
            self.IMFs = IMFs
        try:
            peaks_temp = []
            amps_temp = []
            for imf in range(len(IMFs)):
                p, a = self.compute_peak(IMFs[imf], sf = self.sf, precision = precision, average = 'median')
                peaks_temp.append(p)

                amps_temp.append(a)

            peaks_temp = np.flip(peaks_temp)
            amps_temp = np.flip(amps_temp)
        except:
            pass
        if peaks_function == 'HH1D_max':
            IMFs = EMD_eeg(data)
            IMFs = np.moveaxis(IMFs, 0, 1)
            IP, IF, IA = emd.spectra.frequency_transform(IMFs[:, 1:6], sf, 'nht')
            precision_hh = precision*2
            low = min_freq
            high = max_freq
            steps = int((high-low)/precision_hh)
            edges, bins = emd.spectra.define_hist_bins(low, high, steps, 'log')
            self.IF = np.moveaxis(IF, 0 ,1)
            # Compute the 1d Hilbert-Huang transform (power over carrier frequency)
            spec = emd.spectra.hilberthuang_1d(IF, IA, edges)
            
            spec = np.moveaxis(spec, 0, 1)
            peaks_temp = []
            amps_temp = []
            for e, i in enumerate(spec):
                max_power = np.argmax(i)
                peaks_temp.append(bins[max_power])
                amps_temp.append(spec[e][max_power])
            peaks_temp = np.flip(peaks_temp)
            amps_temp = np.flip(amps_temp)
        #if peaks_function == 'HH1D_weightAVG':

        if peaks_function == 'adapt':
            p, a = self.compute_peaks_raw(data, alphaband, precision = precision, average = 'median')
            FREQ_BANDS = alpha2bands(p)
            peaks_temp, amps_temp = self.compute_peaks_raw(data, FREQ_BANDS, precision = precision, average = 'median')
        if peaks_function == 'fixed':
            peaks_temp, amps_temp = self.compute_peaks_raw(data, FREQ_BANDS, precision = precision, average = 'median')
        if peaks_function == 'harmonic_peaks':
            p, a = self.extract_all_peaks(data, sf, precision, max_freq = sf/2)
            max_n, peaks_temp, amps_temp, self.harmonics, harm_peaks, harm_peaks_fit = harmonic_peaks_fit (p, a, min_freq, max_freq, min_harms = min_harms, harm_limit = harm_limit)
            list_harmonics = functools_reduce(self.harmonics)
            list_harmonics = list(set(abs(np.array(list_harmonics))))
            list_harmonics = [h for h in list_harmonics if h <= harm_limit]
            list_harmonics = np.sort(list_harmonics)
            self.all_harmonics = list_harmonics
            self.n_harmonic_peaks = len(peaks_temp)
        if peaks_function == 'cepstrum':
            cepstrum_, quefrency_vector = cepstrum(self.data, self.sf)
            peaks_temp, amps_temp = cepstral_peaks(cepstrum_, quefrency_vector, 0.4, 0.02)          
            peaks_temp = list(np.flip(peaks_temp))
            peaks_temp = [np.round(p, 2) for p in peaks_temp]
            amps_temp = list(np.flip(amps_temp))
        if peaks_function == 'FOOOF':
            nfft = sf/precision
            nperseg = sf/precision
            freqs1, psd = scipy.signal.welch(data, self.sf, nfft = nfft, nperseg = nperseg)
            self.freqs = freqs1
            self.psd = psd
            fm = FOOOF(peak_width_limits=[precision*2, 3], max_n_peaks=10, min_peak_height=0.2)
            freq_range = [(sf/len(data))*2, 45]
            fm.fit(freqs1, psd, freq_range)
            peaks_temp = []
            amps_temp = []
            for p in range(len(fm.peak_params_)):
                try:
                    peaks_temp.append(fm.peak_params_[p][0])
                    amps_temp.append(fm.peak_params_[p][1])
                except:
                    pass
            peaks_temp = [np.round(p, 2) for p in peaks_temp]
        if peaks_function == 'FOOOF_EEMD':
            nfft = sf/precision
            nperseg = sf/precision
            IMFs = EMD_eeg(data)[1:6]
            self.IMFs = IMFs
            peaks_temp = []
            amps_temp = []
            for imf in IMFs:
                freqs1, psd = scipy.signal.welch(imf, sf, nfft = nfft, nperseg = nperseg)
                self.freqs = freqs1
                self.psd = psd
                fm = FOOOF(peak_width_limits=[precision*2, 3], max_n_peaks=10, min_peak_height=0.2)
                freq_range = [(sf/len(data))*2, 45]
                fm.fit(freqs1, psd, freq_range)
                try:
                    peaks_temp.append(fm.peak_params_[0][0])
                    amps_temp.append(fm.peak_params_[0][1])
                except:
                    pass
            peaks_temp = [np.round(p, 2) for p in peaks_temp]
        peaks = np.array(peaks_temp)
        peaks = np.around(peaks, 3)
        amps = np.array(amps_temp)
        return peaks, amps
    
    '''Listening methods'''
    
    def listen_scale (self, scale, fund = 250, length = 500):
        if scale == 'peaks':
            scale = self.peaks_ratios
        if scale == 'diss':
            try:
                scale = self.diss_scale
            except:
                pass
        if scale == 'HE':
            try:
                scale = list(self.HE_scale)
            except:
                pass
        scale = np.around(scale, 3)
        print('Scale:', scale)
        scale = list(scale)
        scale = [1]+scale
        for s in scale:
            freq = fund*s
            note = make_chord(freq, [1])
            note = np.ascontiguousarray(np.vstack([note,note]).T)
            sound = pygame.sndarray.make_sound(note)
            sound.play(loops=0, maxtime=0, fade_ms=0)
            pygame.time.wait(int(sound.get_length() * length))
            
    '''Generic method to fit all Biotuner methods'''
    
    def fit_all(self, data, compute_diss = True, compute_HE = True, compute_peaks_extension = True):
        biotuning = biotuner(self.sf, peaks_function = self.peaks_function, precision = self.precision, n_harm = self.n_harm)
        biotuning.peaks_extraction(data)
        biotuning.compute_peaks_metrics()
        if compute_diss == True:
            biotuning.compute_diss_curve(input_type = 'peaks', plot = False)
        if compute_peaks_extension == True:
            biotuning.peaks_extension(method = 'multi_consonant_harmonic_fit', harm_function = 'mult', cons_limit = 0.01)
        if compute_HE == True:
            biotuning.compute_harmonic_entropy(input_type = 'extended_peaks', plot_entropy = False)
        return biotuning
    
    def info(self, metrics=False, scales=False, whatever=False):
        if metrics == True:
            print('METRICS')
            print(vars(self))
        
        else:
            print(vars(self))
        return


import numpy as np
from PyEMD import EMD, EEMD
from scipy.signal import butter, lfilter
import matplotlib.pyplot as plt
import seaborn as sbn
from scipy import stats
from numpy import array, zeros, ones, arange, log2, sqrt, diff, concatenate
import secrets
import biotuner
from biotuner.biotuner_functions import *
from biotuner.biotuner_utils import *
from biotuner.biotuner_object import *


'''BIOTUNER 2D'''


def surrogate_signal(data, surr_type = 'pink', low_cut = 0.5, high_cut = 150, sf = 1000):
    if surr_type == 'AAFT':
        indexes = [x for x in range(len(data))]
        data = np.stack((data, indexes))
        data = AAFT_surrogates(Surrogates, data)
        data = data[0]
        data = butter_bandpass_filter(data, low_cut, high_cut, sf, 4)
    if surr_type == 'phase':
        len_data = len(data)
        data = phaseScrambleTS(data)
        data = butter_bandpass_filter(data[0:len_data], low_cut, high_cut, sf, 4)
    if surr_type == 'shuffle':
        np.random.shuffle(data)
    if surr_type == 'white':
        beta = 0 
    if surr_type == 'pink':
        beta  = 1
    if surr_type == 'brown':
        beta = 2
    if surr_type == 'blue':
        beta  = -1
    if surr_type == 'white' or surr_type == 'pink' or surr_type == 'brown' or surr_type == 'blue':
        data = cn.powerlaw_psd_gaussian(beta, len(data))
        data = butter_bandpass_filter(data, low_cut, high_cut, sf, 4)
    return data
    
def surrogate_signal_matrices(data, surr_type = 'pink', low_cut = 0.5, high_cut = 150, sf = 1000):
    if np.ndim(data) == 2:
        for i in range(len(data)):
            data[i] = surrogate_signal(data[i], surr_type = surr_type, low_cut = low_cut, high_cut = high_cut, sf = sf)
    if np.ndim(data) == 3:
        for i in range(len(data)):
            for j in range(len(data[i])):
                data[i][j] = surrogate_signal(data[i][j], surr_type = surr_type, low_cut = low_cut, high_cut = high_cut, sf = sf)
    return data

def compute_peaks_matrices (data, peaks_function = 'EMD', precision = 0.25, sf = 1000, max_freq = 80, save = True, suffix = 'default'):
    if np.ndim(data) == 2:
        peaks_tot = []
        amps_tot = []
        for i in range(len(data)):
            peaks, amps = compute_peaks_ts(data[i], peaks_function = peaks_function, FREQ_BANDS = None, precision = precision, sf = sf, max_freq = max_freq) 
            peaks_tot.append(peaks)
            amps_tot.append(amps)
        peaks_total = np.array(peaks_tot)
        amps_total = np.array(amps_tot)
        if save == True:
            np.save('peaks_{}_{}'.format(peaks_function, suffix), peaks_total)
            np.save('amps_{}_{}'.format(peaks_function, suffix), amps_total)
    if np.ndim(data) == 3:
        peaks_tot = []
        amps_tot = []
        for i in range(len(data)):
            peaks_temp = []
            amps_temp = []
            for j in range(len(data[i])):
                peaks, amps = compute_peaks_ts(data[i][j], peaks_function = peaks_function, FREQ_BANDS = None, precision = precision, sf = sf, max_freq = max_freq)
                peaks_temp.append(peaks)
                amps_temp.append(amps)
            peaks_tot.append(peaks_temp)
            amps_tot.append(amps_temp)
        peaks_total = np.array(peaks_tot)
        amps_total = np.array(amps_tot)
        if save == True:
            np.save('peaks_{}_{}'.format(peaks_function, suffix), peaks_total)
            np.save('amps_{}_{}'.format(peaks_function, suffix), amps_total)
    return peaks_total, amps_total

def compute_peaks_surrogates (data, conditions, peaks_function = 'EMD', precision = 0.25, sf = 1000, max_freq = 80, low_cut = 0.5, high_cut = 150, save = False):
    peaks_tot = []
    amps_tot = []
    for e, c in enumerate(conditions):
        print('Condition (', e+1, 'of', len(conditions), '):', c)
        if c == 'og_data':
            peaks, amps = compute_peaks_matrices(data, peaks_function, precision, sf, max_freq)
        if c != 'og_data':
            data = surrogate_signal_matrices(data, surr_type = c, low_cut = low_cut, high_cut = high_cut, sf = sf)
            peaks, amps = compute_peaks_matrices(data, peaks_function, precision, sf, max_freq, save = False, suffix = 'default')
        peaks_tot.append(peaks)
        amps_tot.append(amps)
    peaks_total = np.array(peaks_tot)
    amps_total = np.array(amps_tot)
        
    return peaks_total, amps_total

def peaks_to_metrics_matrices (peaks, n_harm = 10):
    cons = []
    euler = []
    tenney = []
    harm_fit = []
    metrics_dict = {}
    if np.ndim(peaks) == 3:
        for i in range(len(peaks)):
            cons_temp = []
            euler_temp = []
            tenney_temp = []
            harm_fit_temp = []
            for j in range(len(peaks[i])):
                
                metrics, metrics_list = peaks_to_metrics(peaks[i][j], n_harm)
                cons_temp.append(metrics_list[0])
                euler_temp.append(metrics_list[1])
                tenney_temp.append(metrics_list[2])
                harm_fit_temp.append(metrics_list[3])
            cons.append(cons_temp)
            euler.append(euler_temp)
            tenney.append(tenney_temp)
            harm_fit.append(harm_fit_temp)
        metrics_dict['cons'] = cons
        metrics_dict['euler'] = euler
        metrics_dict['tenney'] = tenney
        metrics_dict['harm_fit'] = harm_fit
    return np.array([cons, euler, tenney, harm_fit]), metrics_dict

def graph_surrogates(data, conditions, metric_to_graph, peaks_function, savefolder, run):
    peaks_avg_tot = []
    metric_tot = []
    for c in conditions:
        if c != 'eeg':
            data_ = surrogate_signal_matrices(data, surr_type = c, low_cut = 0.5, high_cut = 150, sf = 1000)
        else:
            data_ = data
        peaks_avg = []
        metric = []
        for t in range(len(data_)):
            _data_ = data_[t][:]
            biotuning = biotuner(1000, peaks_function = peaks_function, precision = 0.1, n_harm = 10,
                            ratios_n_harms = 10, ratios_inc_fit = False, ratios_inc = False) # Initialize biotuner object
            biotuning.peaks_extraction(_data_, ratios_extension = True, max_freq = 50)
            print(biotuning.peaks)
            biotuning.compute_peaks_metrics()
            peaks_avg.append(np.average(biotuning.peaks))
            metric.append(biotuning.peaks_metrics[metric_to_graph])
        metric_tot.append(metric)
        peaks_avg_tot.append(np.average(peaks_avg))
        print(run)

    graph_dist(metric_tot, metric = metric_to_graph, ref = metric_tot[0], dimensions = [0], labs = conditions, savefolder = savefolder,         subject = '2', run = run, adapt = 'False')
    
    

def graph_dist(dist, metric = 'diss', ref = None, dimensions = [0, 1], labs = ['eeg', 'phase', 'AAFT', 'pink', 'white'], savefolder = '\\', subject = '0', run = '0', adapt = 'False', peak_function = 'EEMD'):
    #print(len(dist), len(dist[0]), len(dist[1]), len(dist[2]), len(dist[3]))
    #if ref == None:
    #    ref = dist[0]
    if metric == 'dissonance':
        m = 'Dissonance (From Sethares (2005))'
    if metric == 'euler':
        m = 'Consonance (Euler <Gradus Suavitatis>)'
    if metric == 'diss_euler':
        m = 'Consonance (Euler <Gradus Suavitatis>) of dissonant minima'
    if metric == 'diss_n_steps':
        m = 'Number of dissonant minima'
    if metric == 'tenney':
        m = 'Tenney Height'
    if metric == 'harmsim':
        m = 'Harmonic similarity of peaks'
    if metric == 'diss_harm_sim':
        m = 'Harmonic similarity of scale'
    if metric == 'harm_fit':
        m = 'Harmonic fitness between peaks'
    if metric == 'cons':
        m = 'Averaged consonance of all paired peaks ratios'
    if metric == 'n_harmonic_peaks':
        m = 'Number of harmonic peaks'

        

    plt.rcParams['axes.facecolor'] = 'black'
    fig = plt.figure(figsize=(14,10))
    colors = ['cyan', 'deeppink', 'white', 'yellow', 'orange'] 
    
    xcoords = []
    
    
    for dim in dimensions:
        labs = labs
        if dim == 0:
            dimension = 'trials'
        if dim == 1:
            dimension = 'channels'
        
        
        for d, color, enum in zip(dist, colors, range(len(dist))):
            #d = d[~np.isnan(d)]
            #print(d.shape)
            if dimensions == [0]:
                sbn.distplot(d, color =color)
                secure_random = secrets.SystemRandom()
                if len(d) < len(ref):
                    ref = secure_random.sample(list(ref), len(d))
                if len(ref) < len(d):
                    d = secure_random.sample(list(d), len(ref))
                t, p = stats.ttest_rel(d, ref)
            if dimensions != [0]:
                sbn.distplot(np.nanmean(d, dim), color =color)
                secure_random = secrets.SystemRandom()
                if len(d) < len(ref):
                    ref = secure_random.sample(list(ref), len(d))
                if len(ref) < len(d):
                    d = secure_random.sample(list(d), len(ref))
                t, p = stats.ttest_rel(np.nanmean(ref, dim), np.nanmean(d, dim))
                
            if p < 0.05:
                labs[enum] = labs[enum]+' *'
                #xcoords.append(np.average(d))

                #for xc in xcoords:
                #    plt.axvline(x=xc, c='white')

        if len(labs) == 2:
            fig.legend(labels=[labs[0], labs[1]], 
                   loc = [0.66, 0.68], fontsize = 16, facecolor = 'white')
        if len(labs) == 3:
            fig.legend(labels=[labs[0], labs[1], labs[2]], 
                   loc = [0.66, 0.68], fontsize = 16, facecolor = 'white')
        if len(labs) == 4:
            fig.legend(labels=[labs[0], labs[1], labs[2], labs[3]], 
                   loc = [0.66, 0.68], fontsize = 16, facecolor = 'white')                    
        plt.xlabel(m, fontsize = '16')
        plt.ylabel('Proportion of samples', fontsize = '16')
        #plt.xlim([0.25, 0.7])
        plt.grid(color='white', linestyle='-.', linewidth=0.7)
        plt.suptitle('Comparing ' + m+ ' \nfor EEG, surrogate data, and noise signals across ' + dimension, fontsize = '22')
        fig.savefig(savefolder+'{}_distribution_s{}-bloc{}_EMD_adapt-{}_{}_{}.png'.format(metric, subject, run, adapt, dimension, peak_function), dpi=300)
        plt.clf()
        
        
def diss_curve_multi (freqs, amps, denom=10, max_ratio=2, bound = 0.1):
    from numpy import array, linspace, empty, concatenate
    from scipy.signal import argrelextrema
    from fractions import Fraction
    plt.figure(figsize=(18, 8))
    diss_minima_tot = []
    for fr, am in zip(freqs, amps):
        freqs = np.array([x*128 for x in fr])
        am = np.interp(am, (np.array(am).min(), np.array(am).max()), (0.3, 0.7))
        r_low = 1
        alpharange = max_ratio
        method = 'min'

        n = 1000
        diss = empty(n)
        a = concatenate((am, am))
        for i, alpha in enumerate(linspace(r_low, alpharange, n)):
            f = concatenate((freqs, alpha*freqs))
            d = dissmeasure(f, a, method)
            diss[i] = d

        
        plt.plot(linspace(r_low, alpharange, len(diss)), diss)
        plt.xscale('log')
        plt.xlim(r_low, alpharange)

        plt.xlabel('frequency ratio')
        plt.ylabel('sensory dissonance')


        diss_minima = argrelextrema(diss, np.less)
        diss_minima_tot.append(list(diss_minima[0]))
        #print(diss_minima)
    
    diss_tot = [item for sublist in diss_minima_tot for item in sublist]
    diss_tot.sort()
    new_minima = []
    
    for i in range(len(diss_tot)-1):
        if (diss_tot[i+1] - diss_tot[i]) < bound:
            new_minima.append((diss_tot[i]+diss_tot[i+1])/2)
    #print(new_minima)
    intervals = []
    for d in range(len(new_minima)):
        #print(new_minima[d])
        frac = Fraction(new_minima[d]/(n/(max_ratio-1))+1).limit_denominator(denom)
        print(frac)
        frac = (frac.numerator, frac.denominator)
        intervals.append(frac)
    
    #intervals = [(123, 100), (147, 100), (159, 100), (9, 5), (2, 1)]
    intervals.append((2, 1))
    #print(intervals)
    for n, d in intervals:
        plt.axvline(n/d, color='silver')
    plt.axvline(1.001, linewidth=1, color='black')
    ax = plt.gca()
    ax.set_facecolor('white')
    ax.axhline(linewidth=1, color='black')
    
    plt.xscale('linear')
    plt.minorticks_off()
    plt.xticks([n/d for n, d in intervals],
               ['{}/{}'.format(n, d) for n, d in intervals], fontsize = 14)
    plt.yticks(fontsize = 14)
    plt.tight_layout()
    plt.show()
    return diss_minima_tot